@@ -25,6 +25,7 @@
 
    cpu		H316/H516 CPU
 
+   06-Nov-04	RMS	Added =n to SHOW HISTORY
    04-Jan-04	RMS	Removed unnecessary compare
    31-Dec-03	RMS	Fixed bug in cpu_set_hist
    24-Oct-03	RMS	Added DMA/DMC support, instruction history
@@ -338,7 +339,7 @@ MTAB cpu_mod[] = {
 		&cpu_set_nchan, &cpu_show_nchan, NULL },
 	{ UNIT_DMC, 0, "no DMC", "NODMC", NULL },
 	{ UNIT_DMC, UNIT_DMC, "DMC", "DMC", NULL },
-	{ MTAB_XTD|MTAB_VDV|MTAB_NMO, 0, "HISTORY", "HISTORY",
+	{ MTAB_XTD|MTAB_VDV|MTAB_NMO|MTAB_SHP, 0, "HISTORY", "HISTORY",
 		&cpu_set_hist, &cpu_show_hist },
 	{ MTAB_XTD | MTAB_VDV | MTAB_NMO, 0, "DMA1", NULL,
 		NULL, &cpu_show_dma, NULL },
@@ -1369,8 +1370,10 @@ return SCPE_OK;
 
 t_stat cpu_show_hist (FILE *st, UNIT *uptr, int32 val, void *desc)
 {
-int32 cr, k, di, op;
+int32 cr, k, di, op, lnt;
+char *cptr = (char *) desc;
 t_value sim_eval;
+t_stat r;
 struct InstHistory *h;
 extern t_stat fprint_sym (FILE *ofile, t_addr addr, t_value *val,
 	UNIT *uptr, int32 sw);
@@ -1378,9 +1381,14 @@ static uint8 has_opnd[16] = {
  0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1, 1, 1 };
 
 if (hst_lnt == 0) return SCPE_NOFNC;			/* enabled? */
+if (cptr) {
+	lnt = (int32) get_uint (cptr, 10, hst_lnt, &r);
+	if ((r != SCPE_OK) || (lnt == 0)) return SCPE_ARG;  }
+else lnt = hst_lnt;
+di = hst_p - lnt;					/* work forward */
+if (di < 0) di = di + hst_lnt;
 fprintf (st, "PC     C A       B       X       ea     IR\n\n");
-di = hst_p;						/* work forward */
-for (k = 0; k < hst_lnt; k++) {				/* print specified */
+for (k = 0; k < lnt; k++) {				/* print specified */
 	h = &hst[(++di) % hst_lnt];			/* entry pointer */
 	if (h->pc & HIST_PC) {				/* instruction? */
 	    cr = (h->pc & HIST_C)? 1: 0;		/* carry */
