@@ -26,6 +26,27 @@
    dp		12557A 2871 disk subsystem
 		13210A 7900 disk subsystem
 
+   07-Oct-04	JDB	Fixed enable/disable from either device
+			Fixed ANY ERROR status for 12557A interface
+			Fixed unattached drive status for 12557A interface
+			Status cmd without prior STC DC now completes (12557A)
+			OTA/OTB CC on 13210A interface also does CLC CC
+			Fixed RAR model
+			Fixed seek check on 13210 if sector out of range
+   20-Aug-04	JDB	Fixes from Dave Bryan
+			- Check status on unattached drive set busy and not ready
+			- Check status tests wrong unit for write protect status
+			- Drive on line sets ATN, will set FLG if polling
+   15-Aug-04	RMS	Controller resumes polling for ATN interrupts after
+			read status (found by Dave Bryan)
+   22-Jul-04	RMS	Controller sets ATN for all commands except
+			read status (found by Dave Bryan)
+   21-Apr-04	RMS	Fixed typo in boot loader (found by Dave Bryan)
+   26-Apr-04	RMS	Fixed SFS x,C and SFC x,C
+			Fixed SR setting in IBL
+			Fixed interpretation of SR<0>
+			Revised IBL loader
+			Implemented DMA SRQ (follows FLG)
    25-Apr-03	RMS	Revised for extended file support
 			Fixed bug(s) in boot (found by Terry Newton)
    10-Nov-02	RMS	Added BOOT command, fixed numerous bugs
@@ -36,6 +57,57 @@
    07-Sep-01	RMS	Moved function prototypes
    29-Nov-00	RMS	Made variable names unique
    21-Nov-00	RMS	Fixed flag, buffer power up state
+
+   The simulator uses a number of state variables:
+
+   dpc_busy		set to drive number + 1 when the controller is busy
+			of the unit in use
+   dpd_xfer		set to 1 if the data channel is executing a data transfer
+   dpd_wval		set to 1 by OTx if either !dpc_busy or dpd_xfer
+   dpc_poll		set to 1 if attention polling is enabled
+
+   dpc_busy and dpd_xfer are set together at the start of a read, write, refine,
+   or init.  When data transfers are complete (CLC DC), dpd_xfer is cleared, but the
+   operation is not necessarily over.  When the operation is complete, dpc_busy
+   is cleared and the command channel flag is set.
+
+   dpc_busy && !dpd_xfer && STC DC (controller is busy, data channel transfer has
+   been terminated by CLC DC, but a word has been placed in the data channel buffer)
+   indicates data overrun.
+
+   dpd_wval is used in write operations to fill out the sector buffer with 0's
+   if only a partial sector has been transferred.
+
+   dpc_poll indicates whether seek completion polling can occur.  It is cleared
+   by reset and CLC CC and set by issuance of a seek or completion of check status.
+
+   The controller's "Record Address Register" (RAR) contains the CHS address of
+   the last Seek or Address Record command executed.  The RAR is shared among
+   all drives on the controller.  In addition, each drive has an internal
+   position register that contains the last cylinder position transferred to the
+   drive during Seek command execution (data operations always start with the
+   RAR head and sector position).
+
+   In a real drive, the address field of the sector under the head is read and
+   compared to the RAR.  When they match, the target sector is under the head
+   and is ready for reading or writing.  If a match doesn't occur, an Address
+   Error is indicated.  In the simulator, the address field is obtained from the
+   drive's current position register during a read, i.e., the "on-disc" address
+   field is assumed to match the current position.
+
+   References:
+   - 7900A Disc Drive Operating and Service Manual (07900-90002, Feb-1975)
+   - 13210A Disc Drive Interface Kit Operating and Service Manual
+            (13210-90003, Nov-1974)
+   - 12557A Cartridge Disc Interface Kit Operating and Service Manual
+            (12557-90001, Sep-1970)
+
+   The following implemented behaviors have been inferred from secondary sources
+   (diagnostics, operating system drivers, etc.), due to absent or contradictory
+   authoritative information; future correction may be needed:
+
+     1. Status bit 15 (ATTENTION) does not set bit 0 (ANY ERROR) on the 12557A.
+     2. Omitting STC DC before Status Check does not set DC flag but does poll.
 */
 
 #include "hp2100_defs.h"
@@ -43,7 +115,7 @@
 #define UNIT_V_WLK	(UNIT_V_UF + 0)			/* write locked */
 #define UNIT_WLK	(1 << UNIT_V_WLK)
 #define FNC		u3				/* saved function */
-#define CYL		u4				/* cylinder */
+#define DRV		u4				/* drive number (DC) */
 #define UNIT_WPRT	(UNIT_WLK | UNIT_RO)		/* write prot */
 
 #define DP_N_NUMWD	7
@@ -115,19 +187,27 @@
 #define STA_BSY		0000004				/* seeking */
 #define STA_DTE		0000002				/* data error */
 #define STA_ERR		0000001				/* any error (d) */
-#define STA_ALLERR	(STA_ATN + STA_1ST + STA_OVR + STA_RWU + STA_ACU + \
-			 STA_SKI + STA_SKE + STA_NRDY + STA_EOC + STA_AER + \
-			 STA_FLG + STA_BSY + STA_DTE)
-#define STA_MBZ13	(STA_ATN + STA_RWU + STA_SKI)	/* zero in 13210 */
+
+#define STA_ERSET2	(STA_1ST | STA_OVR | STA_RWU | STA_ACU | \
+			 STA_SKI | STA_SKE | STA_NRDY | \
+			 STA_EOC | STA_AER | STA_DTE)	/* 12557A error set */
+#define STA_ERSET3	(STA_ATN | STA_1ST | STA_OVR | STA_RWU | STA_ACU | \
+			 STA_SKI | STA_SKE | STA_NRDY | STA_EOC | STA_AER | \
+			 STA_FLG | STA_BSY | STA_DTE)	/* 13210A error set */
+#define STA_ANYERR	(dp_ctype ? STA_ERSET3 : STA_ERSET2)
+
+#define STA_UNLOADED	(dp_ctype ? (STA_NRDY | STA_BSY) : STA_NRDY)
+#define STA_MBZ13	(STA_ATN | STA_RWU | STA_SKI)	/* zero in 13210 */
 
 extern uint16 *M;
 extern uint32 PC, SR;
-extern uint32 dev_cmd[2], dev_ctl[2], dev_flg[2], dev_fbf[2];
+extern uint32 dev_cmd[2], dev_ctl[2], dev_flg[2], dev_fbf[2], dev_srq[2];
 extern int32 sim_switches;
 extern UNIT cpu_unit;
 
 int32 dp_ctype = 1;					/* ctrl type */
 int32 dpc_busy = 0;					/* cch unit */
+int32 dpc_poll = 0;					/* cch poll enable */
 int32 dpc_cnt = 0;					/* check count */
 int32 dpc_eoc = 0;					/* end of cyl */
 int32 dpc_stime = 100;					/* seek time */
@@ -139,9 +219,10 @@ int32 dpc_obuf = 0;					/* cch buffers */
 int32 dpd_xfer = 0;					/* xfer in prog */
 int32 dpd_wval = 0;					/* write data valid */
 int32 dp_ptr = 0;					/* buffer ptr */
-uint8 dpc_rarc[DP_NUMDRV] = { 0 };			/* cylinder */
-uint8 dpc_rarh[DP_NUMDRV] = { 0 };			/* head */
-uint8 dpc_rars[DP_NUMDRV] = { 0 };			/* sector */
+uint8 dpc_rarc = 0;					/* RAR cylinder */
+uint8 dpc_rarh = 0;					/* RAR head */
+uint8 dpc_rars = 0;					/* RAR sector */
+uint8 dpc_ucyl[DP_NUMDRV] = { 0 };			/* unit cylinder */
 uint16 dpc_sta[DP_NUMDRV] = { 0 };			/* status regs */
 uint16 dpxb[DP_NUMWD];					/* sector buffer */
 
@@ -167,8 +248,8 @@ t_stat dp_showtype (FILE *st, UNIT *uptr, int32 val, void *desc);
 */
 
 DIB dp_dib[] = {
-	{ DPD, 0, 0, 0, 0, &dpdio },
-	{ DPC, 0, 0, 0, 0, &dpcio }  };
+	{ DPD, 0, 0, 0, 0, 0, &dpdio },
+	{ DPC, 0, 0, 0, 0, 0, &dpcio }  };
 
 #define dpd_dib dp_dib[0]
 #define dpc_dib dp_dib[1]
@@ -178,14 +259,15 @@ UNIT dpd_unit = { UDATA (&dpd_svc, 0, 0) };
 REG dpd_reg[] = {
 	{ ORDATA (IBUF, dpd_ibuf, 16) },
 	{ ORDATA (OBUF, dpd_obuf, 16) },
+	{ BRDATA (DBUF, dpxb, 8, 16, DP_NUMWD) },
+	{ DRDATA (BPTR, dp_ptr, DP_N_NUMWD) },
 	{ FLDATA (CMD, dpd_dib.cmd, 0) },
 	{ FLDATA (CTL, dpd_dib.ctl, 0) },
 	{ FLDATA (FLG, dpd_dib.flg, 0) },
 	{ FLDATA (FBF, dpd_dib.fbf, 0) },
+	{ FLDATA (SRQ, dpd_dib.srq, 0) },
 	{ FLDATA (XFER, dpd_xfer, 0) },
 	{ FLDATA (WVAL, dpd_wval, 0) },
-	{ BRDATA (DBUF, dpxb, 8, 16, DP_NUMWD) },
-	{ DRDATA (BPTR, dp_ptr, DP_N_NUMWD) },
 	{ ORDATA (DEVNO, dpd_dib.devno, 6), REG_HRO },
 	{ NULL }  };
 
@@ -199,7 +281,7 @@ DEVICE dpd_dev = {
 	1, 10, DP_N_NUMWD, 1, 8, 16,
 	NULL, NULL, &dpc_reset,
 	NULL, NULL, NULL,
-	&dpd_dib, 0 };
+	&dpd_dib, DEV_DISABLE };
 
 /* DPC data structures
 
@@ -221,24 +303,25 @@ UNIT dpc_unit[] = {
 
 REG dpc_reg[] = {
 	{ ORDATA (OBUF, dpc_obuf, 16) },
-	{ ORDATA (BUSY, dpc_busy, 3), REG_RO },
+	{ ORDATA (BUSY, dpc_busy, 4), REG_RO },
 	{ ORDATA (CNT, dpc_cnt, 5) },
 	{ FLDATA (CMD, dpc_dib.cmd, 0) },
 	{ FLDATA (CTL, dpc_dib.ctl, 0) },
 	{ FLDATA (FLG, dpc_dib.flg, 0) },
 	{ FLDATA (FBF, dpc_dib.fbf, 0) },
+	{ FLDATA (SRQ, dpc_dib.srq, 0) },
 	{ FLDATA (EOC, dpc_eoc, 0) },
-	{ BRDATA (RARC, dpc_rarc, 8, 8, DP_NUMDRV) },
-	{ BRDATA (RARH, dpc_rarh, 8, 2, DP_NUMDRV) },
-	{ BRDATA (RARS, dpc_rars, 8, 4, DP_NUMDRV) },
+	{ FLDATA (POLL, dpc_poll, 0) },
+	{ DRDATA (RARC, dpc_rarc, 8), PV_RZRO },
+	{ DRDATA (RARH, dpc_rarh, 2), PV_RZRO },
+	{ DRDATA (RARS, dpc_rars, 5), PV_RZRO },
+	{ BRDATA (CYL, dpc_ucyl, 10, 8, DP_NUMDRV), PV_RZRO },
 	{ BRDATA (STA, dpc_sta, 8, 16, DP_NUMDRV) },
 	{ DRDATA (CTIME, dpc_ctime, 24), PV_LEFT },
 	{ DRDATA (DTIME, dpc_dtime, 24), PV_LEFT },
 	{ DRDATA (STIME, dpc_stime, 24), PV_LEFT },
-	{ DRDATA (XTIME, dpc_xtime, 24), REG_NZ + PV_LEFT },
+	{ DRDATA (XTIME, dpc_xtime, 24), REG_NZ | PV_LEFT },
 	{ FLDATA (CTYPE, dp_ctype, 0), REG_HRO },
-	{ URDATA (UCYL, dpc_unit[0].CYL, 10, 8, 0,
-		  DP_NUMDRV, PV_LEFT | REG_HRO) },
 	{ URDATA (UFNC, dpc_unit[0].FNC, 8, 8, 0,
 		  DP_NUMDRV, REG_HRO) },
 	{ URDATA (CAPAC, dpc_unit[0].capac, 10, T_ADDR_W, 0,
@@ -275,14 +358,14 @@ int32 devd;
 devd = IR & I_DEVMASK;					/* get device no */
 switch (inst) {						/* case on opcode */
 case ioFLG:						/* flag clear/set */
-	if ((IR & I_HC) == 0) { setFLG (devd); }	/* STF */
+	if ((IR & I_HC) == 0) { setFSR (devd); }	/* STF */
 	break;
 case ioSFC:						/* skip flag clear */
 	if (FLG (devd) == 0) PC = (PC + 1) & VAMASK;
-	return dat;
+	break;
 case ioSFS:						/* skip flag set */
 	if (FLG (devd) != 0) PC = (PC + 1) & VAMASK;
-	return dat;
+	break;
 case ioOTX:						/* output */
 	dpd_obuf = dat;
 	if (!dpc_busy || dpd_xfer) dpd_wval = 1;	/* if !overrun, valid */
@@ -306,7 +389,7 @@ case ioCTL:						/* control clear/set */
 	break;
 default:
 	break;  }
-if (IR & I_HC) { clrFLG (devd); }			/* H/C option */
+if (IR & I_HC) { clrFSR (devd); }			/* H/C option */
 return dat;
 }
 
@@ -318,16 +401,13 @@ int32 devd = dpd_dib.devno;
 devc = IR & I_DEVMASK;					/* get device no */
 switch (inst) {						/* case on opcode */
 case ioFLG:						/* flag clear/set */
-	if ((IR & I_HC) == 0) { setFLG (devc); }	/* STF */
+	if ((IR & I_HC) == 0) { setFSR (devc); }	/* STF */
 	break;
 case ioSFC:						/* skip flag clear */
 	if (FLG (devc) == 0) PC = (PC + 1) & VAMASK;
-	return dat;
+	break;
 case ioSFS:						/* skip flag set */
 	if (FLG (devc) != 0) PC = (PC + 1) & VAMASK;
-	return dat;
-case ioOTX:						/* output */
-	dpc_obuf = dat;
 	break;
 case ioLIX:						/* load */
 	dat = 0;
@@ -335,6 +415,10 @@ case ioMIX:						/* merge */
 	for (i = 0; i < DP_NUMDRV; i++)
 	    if (dpc_sta[i] & STA_ATN) dat = dat | (1 << i);
 	break;
+case ioOTX:						/* output */
+	dpc_obuf = dat;
+	if (!dp_ctype) break;
+	IR = IR | I_CTL;				/* 13210 OTx causes CLC */
 case ioCTL:						/* control clear/set */
 	if (IR & I_CTL) {				/* CLC? */
 	    clrCTL (devc);				/* clr cmd, ctl */
@@ -342,7 +426,8 @@ case ioCTL:						/* control clear/set */
 	    if (dpc_busy) sim_cancel (&dpc_unit[dpc_busy - 1]);
 	    sim_cancel (&dpd_unit);			/* cancel dch */
 	    dpd_xfer = 0;				/* clr dch xfer */
-	    dpc_busy = 0;  }				/* clr cch busy */
+	    dpc_busy = 0;				/* clr cch busy */
+	    dpc_poll = 0;  }				/* clr cch poll */
 	else {						/* STC */
 	    setCTL (devc);				/* set ctl */
 	    if (!CMD (devc)) {				/* is cmd clr? */
@@ -350,9 +435,13 @@ case ioCTL:						/* control clear/set */
 		drv = CW_GETDRV (dpc_obuf);		/* get fnc, drv */
 		fnc = CW_GETFNC (dpc_obuf);		/* from cmd word */
 		switch (fnc) {				/* case on fnc */
+		case FNC_SEEK:				/* seek */
+		    dpc_poll = 1;			/* enable polling */
+		    dp_god (fnc, drv, dpc_dtime);	/* sched dch xfr */
+		    break;
 		case FNC_STA:				/* rd sta */
-		    if (dp_ctype) { clrFLG (devd); }	/* 13210? clr dch flag */
-		case FNC_SEEK: case FNC_CHK:		/* seek, check */
+		    if (dp_ctype) { clrFSR (devd); }	/* 13210? clr dch flag */
+		case FNC_CHK:				/* check */
 		case FNC_AR:				/* addr rec */
 		    dp_god (fnc, drv, dpc_dtime);	/* sched dch xfr */
 		    break;
@@ -365,7 +454,7 @@ case ioCTL:						/* control clear/set */
 	break;
 default:
 	break;  }
-if (IR & I_HC) { clrFLG (devc); }			/* H/C option */
+if (IR & I_HC) { clrFSR (devc); }			/* H/C option */
 return dat;
 }
 
@@ -373,7 +462,7 @@ return dat;
 
 void dp_god (int32 fnc, int32 drv, int32 time)
 {
-dpd_unit.CYL = drv;					/* save unit */
+dpd_unit.DRV = drv;					/* save unit */
 dpd_unit.FNC = fnc;					/* save function */
 sim_activate (&dpd_unit, time);
 return;
@@ -383,15 +472,18 @@ return;
 
 void dp_goc (int32 fnc, int32 drv, int32 time)
 {
-if (sim_is_active (&dpc_unit[drv])) {			/* still seeking? */
+int32 t;
+
+if (t = sim_is_active (&dpc_unit[drv])) {		/* still seeking? */
 	sim_cancel (&dpc_unit[drv]);			/* stop seek */
 	dpc_sta[drv] = dpc_sta[drv] & ~STA_BSY;		/* clear busy */
-	time = time + dpc_stime;  }			/* take longer */
+	time = time + t;  }				/* include seek time */
 dp_ptr = 0;						/* init buf ptr */
 dpc_eoc = 0;						/* clear end cyl */
 dpc_busy = drv + 1;					/* set busy */
 dpd_xfer = 1;						/* xfer in prog */
 dpc_unit[drv].FNC = fnc;				/* save function */
+dpc_sta[drv] = dpc_sta[drv] & ~STA_ATN;			/* clear ATN */
 sim_activate (&dpc_unit[drv], time);			/* activate unit */
 return;
 }
@@ -401,7 +493,7 @@ return;
    This routine handles the data channel transfers.  It also handles
    data transfers that are blocked by seek in progress.
 
-   uptr->CYL	=	target drive
+   uptr->DRV	=	target drive
    uptr->FNC	=	target function
 
    Seek substates
@@ -417,89 +509,85 @@ return;
 
 t_stat dpd_svc (UNIT *uptr)
 {
-int32 drv, devc, devd, st;
+int32 i, drv, devc, devd, st;
 
-drv = uptr->CYL;					/* get drive no */
+drv = uptr->DRV;					/* get drive no */
 devc = dpc_dib.devno;					/* get cch devno */
 devd = dpd_dib.devno;					/* get dch devno */
 switch (uptr->FNC) {					/* case function */
 
+case FNC_AR:						/* arec, need cyl */
 case FNC_SEEK:						/* seek, need cyl */
 	if (CMD (devd)) {				/* dch active? */
-	    dpc_rarc[drv] = DA_GETCYL (dpd_obuf);	/* take cyl word */
+	    dpc_rarc = DA_GETCYL (dpd_obuf);		/* set RAR from cyl word */
 	    dpd_wval = 0;				/* clr data valid */
-	    setFLG (devd);				/* set dch flg */
+	    setFSR (devd);				/* set dch flg */
 	    clrCMD (devd);				/* clr dch cmd */
-	    uptr->FNC = FNC_SEEK1;  }			/* advance state */
+	    if (uptr->FNC == FNC_AR) uptr->FNC = FNC_AR1;
+	    else uptr->FNC = FNC_SEEK1;  }		/* advance state */
 	sim_activate (uptr, dpc_xtime);			/* no, wait more */
 	break;
+
+case FNC_AR1:						/* arec, need hd/sec */
 case FNC_SEEK1:						/* seek, need hd/sec */
 	if (CMD (devd)) {				/* dch active? */
-	    dpc_rarh[drv] = DA_GETHD (dpd_obuf);	/* get head */
-	    dpc_rars[drv] = DA_GETSC (dpd_obuf);	/* get sector */
+	    dpc_rarh = DA_GETHD (dpd_obuf);		/* set RAR from head */
+	    dpc_rars = DA_GETSC (dpd_obuf);		/* set RAR from sector */
 	    dpd_wval = 0;				/* clr data valid */
-	    setFLG (devd);				/* set dch flg */
+	    setFSR (devd);				/* set dch flg */
 	    clrCMD (devd);				/* clr dch cmd */
+	    if (uptr->FNC == FNC_AR1) {
+		setFSR (devc);				/* set cch flg */
+		clrCMD (devc);				/* clr cch cmd */
+		dpc_sta[drv] = dpc_sta[drv] | STA_ATN;	/* set drv attn */
+		break;  }				/* done if Address Record */
 	    if (sim_is_active (&dpc_unit[drv])) {	/* if busy, */
-		dpc_sta[drv] = dpc_sta[drv] | STA_SKE;
-		break;  }				/* error, ignore */
-	    st = abs (dpc_rarc[drv] - dpc_unit[drv].CYL) * dpc_stime;
+		dpc_sta[drv] = dpc_sta[drv] | STA_SKE;	/* seek check */
+		break;  }				/* allow prev seek to cmpl */
+	    if ((dpc_rarc >= DP_NUMCY) || 		/* invalid cyl? */
+	    	(dp_ctype && (dpc_rars >= DP_NUMSC3))) {  /* invalid sector? (13210A) */
+		dpc_sta[drv] = dpc_sta[drv] | STA_SKE;	/* seek check */
+	    	sim_activate (&dpc_unit[drv], 1);	/* schedule drive no-wait */
+		dpc_unit[drv].FNC = FNC_SEEK3;		/* do immed compl w/poll */
+		break;  }
+	    st = abs (dpc_rarc - dpc_ucyl[drv]) * dpc_stime;
 	    if (st == 0) st = dpc_stime;		/* min time */
+	    dpc_ucyl[drv] = dpc_rarc;			/* transfer RAR */
 	    sim_activate (&dpc_unit[drv], st);		/* schedule drive */
 	    dpc_sta[drv] = (dpc_sta[drv] | STA_BSY) &
 		~(STA_SKE | STA_SKI | STA_HUNT);
-	    dpc_unit[drv].CYL = dpc_rarc[drv];		/* on cylinder */
 	    dpc_unit[drv].FNC = FNC_SEEK2;  }		/* set operation */
 	else sim_activate (uptr, dpc_xtime);		/* no, wait more */
 	break;
 
-case FNC_AR:						/* arec, need cyl */
-	if (CMD (devd)) {				/* dch active? */
-	    dpc_rarc[drv] = DA_GETCYL (dpd_obuf);	/* take cyl word */
-	    dpd_wval = 0;				/* clr data valid */
-	    setFLG (devd);				/* set dch flg */
-	    clrCMD (devd);				/* clr dch cmd */
-	    uptr->FNC = FNC_AR1;  }			/* advance state */
-	sim_activate (uptr, dpc_xtime);			/* no, wait more */
-	break;
-case FNC_AR1:						/* arec, need hd/sec */
-	if (CMD (devd)) {				/* dch active? */
-	    dpc_rarh[drv] = DA_GETHD (dpd_obuf);	/* get head */
-	    dpc_rars[drv] = DA_GETSC (dpd_obuf);	/* get sector */
-	    dpd_wval = 0;				/* clr data valid */
-	    setFLG (devc);				/* set cch flg */
-	    clrCMD (devc);				/* clr cch cmd */
-	    setFLG (devd);				/* set dch flg */
-	    clrCMD (devd);  }				/* clr dch cmd */
-	else sim_activate (uptr, dpc_xtime);		/* no, wait more */
-	break;
-
 case FNC_STA:						/* read status */
 	if (CMD (devd) || dp_ctype) {			/* dch act or 13210? */
 	    if (dpc_unit[drv].flags & UNIT_ATT) {	/* attached? */
 		dpd_ibuf = dpc_sta[drv] & ~STA_ERR;	/* clear err */
 		if (dp_ctype) dpd_ibuf =		/* 13210? */
 		    (dpd_ibuf & ~(STA_MBZ13 | STA_PROT)) |
-		    (uptr->flags & UNIT_WPRT? STA_PROT: 0);  }
-	    else dpd_ibuf = STA_NRDY;			/* not ready */
-	    if (dpd_ibuf & STA_ALLERR)			/* errors? set flg */
+		    (dpc_unit[drv].flags & UNIT_WPRT? STA_PROT: 0);  }
+	    else dpd_ibuf = STA_UNLOADED;		/* not ready */
+	    if (dpd_ibuf & STA_ANYERR)			/* errors? set flg */
 		dpd_ibuf = dpd_ibuf | STA_ERR;
-	    setFLG (devd);				/* set dch flg */
+	    setFSR (devd);				/* set dch flg */
 	    clrCMD (devd);				/* clr dch cmd */
-	    clrCMD (devc);				/* clr cch cmd */
-	    dpc_sta[drv] = dpc_sta[drv] &		/* clr sta flags */
-		~(STA_ATN | STA_1ST | STA_OVR |
-		STA_RWU | STA_ACU | STA_EOC |
-		STA_AER | STA_FLG | STA_DTE);  }
-	else sim_activate (uptr, dpc_xtime);		/* wait more */
+	    clrCMD (devc);  }				/* clr cch cmd */
+	dpc_sta[drv] = dpc_sta[drv] &			/* clr sta flags */
+	    ~(STA_ATN | STA_1ST | STA_OVR |
+	    STA_RWU | STA_ACU | STA_EOC |
+	    STA_AER | STA_FLG | STA_DTE);
+	dpc_poll = 1;					/* enable polling */
+	for (i = 0; i < DP_NUMDRV; i++) {		/* loop thru drives */
+	    if (dpc_sta[i] & STA_ATN) {			/* any ATN set? */
+		setFSR (devc);				/* set cch flg */
+		break;  }  }
 	break;
 
 case FNC_CHK:						/* check, need cnt */
 	if (CMD (devd)) {				/* dch active? */
 	    dpc_cnt = dpd_obuf & DA_CKMASK;		/* get count */
 	    dpd_wval = 0;				/* clr data valid */
-/*	    setFLG (devd);				/* set dch flg */
-/*	    clrCMD (devd);				/* clr dch cmd */
 	    dp_goc (FNC_CHK1, drv, dpc_xtime);  }	/* sched drv */
 	else sim_activate (uptr, dpc_xtime);		/* wait more */
 	break;
@@ -535,25 +623,22 @@ drv = uptr - dpc_dev.units;				/* get drive no */
 devc = dpc_dib.devno;					/* get cch devno */
 devd = dpd_dib.devno;					/* get dch devno */
 if ((uptr->flags & UNIT_ATT) == 0) {			/* not attached? */
-	setFLG (devc);					/* set cch flg */
+	setFSR (devc);					/* set cch flg */
 	clrCMD (devc);					/* clr cch cmd */
 	dpc_sta[drv] = 0;				/* clr status */
 	dpc_busy = 0;					/* ctlr is free */
-	dpd_xfer = dpd_wval = 0;
+	dpc_poll = 0;					/* polling disabled */
+	dpd_xfer = 0;
+	dpd_wval = 0;
 	return SCPE_OK;  }
 switch (uptr->FNC) {					/* case function */
 
-case FNC_SEEK2:						/* seek done */
-	dpc_sta[drv] = (dpc_sta[drv] | STA_ATN) & ~STA_BSY;
-	if (uptr->CYL >= DP_NUMCY) {			/* invalid cyl? */
-	    dpc_sta[drv] = dpc_sta[drv] | STA_SKE;
-	    uptr->CYL = DP_NUMCY - 1;  }
-case FNC_SEEK3:						/* waiting for flag */
-	if (dpc_busy || FLG (devc)) {			/* ctrl busy? wait */
-	    uptr->FNC = FNC_SEEK3;			/* next state */
-	    sim_activate (uptr, dpc_xtime);  }
-	else {
-	    setFLG (devc);				/* set cch flg */
+case FNC_SEEK2:						/* positioning done */
+	dpc_sta[drv] = (dpc_sta[drv] | STA_ATN) & ~STA_BSY;  /* fall into cmpl */
+
+case FNC_SEEK3:						/* seek complete */
+	if (dpc_poll) {					/* polling enabled? */
+	    setFSR (devc);				/* set cch flg */
 	    clrCMD (devc);  }				/* clear cmd */
 	return SCPE_OK;
 
@@ -564,20 +649,19 @@ case FNC_RD:						/* read */
 case FNC_CHK1:						/* check */
 	if (dp_ptr == 0) {				/* new sector? */
 	    if (!CMD (devd) && (uptr->FNC != FNC_CHK1)) break;
-	    if (uptr->CYL != dpc_rarc[drv])		/* wrong cyl? */
+	    if (dpc_rarc != dpc_ucyl[drv])		/* RAR cyl miscompare? */
 		dpc_sta[drv] = dpc_sta[drv] | STA_AER;	/* set flag, read */
-	    if (dpc_rars[drv] >= DP_NUMSC) {		/* bad sector? */
+	    if (dpc_rars >= DP_NUMSC) {			/* bad sector? */
 		dpc_sta[drv] = dpc_sta[drv] | STA_AER;	/* set flag, stop */
 		break;  }
 	    if (dpc_eoc) {				/* end of cyl? */
 		dpc_sta[drv] = dpc_sta[drv] | STA_EOC;
 		break;  }
-	    da = GETDA (dpc_rarc[drv], dpc_rarh[drv], dpc_rars[drv]);
-	    dpc_rars[drv] = dpc_rars[drv] + 1;		/* incr address */
-	    if (dpc_rars[drv] >= DP_NUMSC) {		/* end of surf? */
-		dpc_rars[drv] = 0;			/* wrap to */
-		dpc_rarh[drv] = dpc_rarh[drv] ^ 1;	/* next head */
-		dpc_eoc = ((dpc_rarh[drv] & 1) == 0);  }/* calc eoc */
+	    da = GETDA (dpc_rarc, dpc_rarh, dpc_rars);	/* calc disk addr */
+	    dpc_rars = (dpc_rars + 1) % DP_NUMSC;	/* incr sector */
+	    if (dpc_rars == 0) {			/* wrap? */
+  		dpc_rarh = dpc_rarh ^ 1;		/* incr head */
+		dpc_eoc = ((dpc_rarh & 1) == 0);  }	/* calc eoc */
 	    if (err = fseek (uptr->fileref, da * sizeof (int16),
 		SEEK_SET)) break;
 	    fxread (dpxb, sizeof (int16), DP_NUMWD, uptr->fileref);
@@ -589,7 +673,7 @@ case FNC_CHK1:						/* check */
 		if (dpc_cnt == 0) break;  }		/* stop at zero */
 	    dp_ptr = 0;  }				/* wrap buf ptr */
 	if (CMD (devd) && dpd_xfer) {			/* dch on, xfer? */
-	    setFLG (devd); }				/* set flag */
+	    setFSR (devd); }				/* set flag */
 	clrCMD (devd);					/* clr dch cmd */
 	sim_activate (uptr, dpc_xtime);			/* sched next word */
 	return SCPE_OK;
@@ -601,9 +685,9 @@ case FNC_WD:						/* write */
 	    if (uptr->flags & UNIT_WPRT) {		/* wr prot? */
 		dpc_sta[drv] = dpc_sta[drv] | STA_FLG;	/* set status */
 		break;  }				/* done */
-	    if ((uptr->CYL != dpc_rarc[drv]) ||		/* wrong cyl or */
-		(dpc_rars[drv] >= DP_NUMSC)) {		/* bad sector? */
-		dpc_sta[drv] = dpc_sta[drv] | STA_AER;	/* set flag, stop */
+	    if ((dpc_rarc != dpc_ucyl[drv]) ||		/* RAR cyl miscompare? */
+		(dpc_rars >= DP_NUMSC)) {		/* bad sector? */
+		dpc_sta[drv] = dpc_sta[drv] | STA_AER;	/* address error */
 		break;  }
 	    if (dpc_eoc) {				/* end of cyl? */
 	        dpc_sta[drv] = dpc_sta[drv] | STA_EOC;	/* set status */
@@ -611,19 +695,18 @@ case FNC_WD:						/* write */
 	dpxb[dp_ptr++] = dpd_wval? dpd_obuf: 0;		/* store word/fill */
 	dpd_wval = 0;					/* clr data valid */
 	if (dp_ptr >= DP_NUMWD) {			/* buffer full? */
-	    da = GETDA (dpc_rarc[drv], dpc_rarh[drv], dpc_rars[drv]);
-	    dpc_rars[drv] = dpc_rars[drv] + 1;		/* incr address */
-	    if (dpc_rars[drv] >= DP_NUMSC) {		/* end of surf? */
-		dpc_rars[drv] = 0;			/* wrap to */
-		dpc_rarh[drv] = dpc_rarh[drv] ^ 1;	/* next head */
-		dpc_eoc = ((dpc_rarh[drv] & 1) == 0);  }/* calc eoc */
+	    da = GETDA (dpc_rarc, dpc_rarh, dpc_rars);	/* calc disk addr */
+	    dpc_rars = (dpc_rars + 1) % DP_NUMSC;	/* incr sector */
+	    if (dpc_rars == 0) {			/* wrap? */
+		dpc_rarh = dpc_rarh ^ 1;		/* incr head */
+		dpc_eoc = ((dpc_rarh & 1) == 0);  }	/* calc eoc */
 	    if (err = fseek (uptr->fileref, da * sizeof (int16),
 		SEEK_SET)) break;
 	    fxwrite (dpxb, sizeof (int16), DP_NUMWD, uptr->fileref);
 	    if (err = ferror (uptr->fileref)) break;	/* error? */
 	    dp_ptr = 0;  }				/* next sector */
 	if (CMD (devd) && dpd_xfer) {			/* dch on, xfer? */
-	    setFLG (devd); }				/* set flag */
+	    setFSR (devd); }				/* set flag */
 	clrCMD (devd);					/* clr dch cmd */
 	sim_activate (uptr, dpc_xtime);			/* sched next word */
 	return SCPE_OK;
@@ -631,8 +714,8 @@ case FNC_WD:						/* write */
 default:
 	return SCPE_IERR;  }				/* end case fnc */
 
-if (!dp_ctype) dpc_sta[drv] = dpc_sta[drv] | STA_ATN;	/* 12559 sets ATN */
-setFLG (devc);						/* set cch flg */
+dpc_sta[drv] = dpc_sta[drv] | STA_ATN;			/* set ATN */
+setFSR (devc);						/* set cch flg */
 clrCMD (devc);						/* clr cch cmd */
 dpc_busy = 0;						/* ctlr is free */
 dpd_xfer = dpd_wval = 0;
@@ -647,27 +730,30 @@ return SCPE_OK;
 
 t_stat dpc_reset (DEVICE *dptr)
 {
-int32 i;
+int32 drv;
 
-hp_enbdis_pair (&dpc_dev, &dpd_dev);			/* make pair cons */
+hp_enbdis_pair (dptr,					/* make pair cons */
+	(dptr == &dpd_dev)? &dpc_dev: &dpd_dev);
 dpd_ibuf = dpd_obuf = 0;				/* clear buffers */
 dpc_busy = dpc_obuf = 0;
 dpc_eoc = 0;
+dpc_poll = 0;
 dpd_xfer = dpd_wval = 0;
 dp_ptr = 0;
+dpc_rarc = dpc_rarh = dpc_rars = 0;			/* clear RAR */
 dpc_dib.cmd = dpd_dib.cmd = 0;				/* clear cmd */
 dpc_dib.ctl = dpd_dib.ctl = 0;				/* clear ctl */
 dpc_dib.fbf = dpd_dib.fbf = 1;				/* set fbf */
 dpc_dib.flg = dpd_dib.flg = 1;				/* set flg */
+dpc_dib.srq = dpd_dib.flg = 1;				/* srq follows flg */
 sim_cancel (&dpd_unit);					/* cancel dch */
-for (i = 0; i < DP_NUMDRV; i++) {			/* loop thru drives */
-	sim_cancel (&dpc_unit[i]);			/* cancel activity */
-	dpc_unit[i].FNC = 0;				/* clear function */
-	dpc_unit[i].CYL = 0;
-	dpc_rarc[i] = dpc_rarh[i] = dpc_rars[i] = 0;
-	if (dpc_unit[i].flags & UNIT_ATT)
-	    dpc_sta[i] = dpc_sta[i] & STA_1ST;
-	else dpc_sta[i] = 0;  }
+for (drv = 0; drv < DP_NUMDRV; drv++) {			/* loop thru drives */
+	sim_cancel (&dpc_unit[drv]);			/* cancel activity */
+	dpc_unit[drv].FNC = 0;				/* clear function */
+	dpc_ucyl[drv] = 0;				/* clear drive pos */
+	if (dpc_unit[drv].flags & UNIT_ATT)
+	    dpc_sta[drv] = dpc_sta[drv] & STA_1ST;	/* first seek status */
+	else dpc_sta[drv] = 0;  }			/* clear status */
 return SCPE_OK;
 }
 
@@ -681,7 +767,11 @@ t_stat r;
 drv = uptr - dpc_dev.units;				/* get drive no */
 r = attach_unit (uptr, cptr);				/* attach unit */
 if (r != SCPE_OK) return r;
-dpc_sta[drv] = dpc_sta[drv] | STA_1ST;			/* update status */
+dpc_sta[drv] = dpc_sta[drv] | STA_ATN | STA_1ST;	/* update status */
+if (dpc_poll) {						/* polling enabled? */
+	dpc_dib.fbf = 1;				/* set fbf */
+	dpc_dib.flg = 1;				/* set flg */
+	dpc_dib.srq = 1; }				/* srq follows flg */
 return r;
 }
 
@@ -711,27 +801,24 @@ return SCPE_OK;
 
 /* 7900/7901 bootstrap routine (HP 12992F ROM) */
 
-#define LDR_BASE	077
-#define CHANGE_DEV	(1 << 24)
-
-static const int32 dboot[IBL_LNT] = {
-	0106700+CHANGE_DEV,	/*ST CLC DC		; clr dch */
-	0106701+CHANGE_DEV,	/*   CLC CC		; clr cch */
+const uint16 dp_rom[IBL_LNT] = {
+	0106710,		/*ST CLC DC		; clr dch */
+	0106711,		/*   CLC CC		; clr cch */
 	0017757,		/*   JSB STAT		; get status */
 	0067746,		/*SK LDB SKCMD		; seek cmd */
-	0106600+CHANGE_DEV,	/*   OTB DC		; cyl # */
-	0103700+CHANGE_DEV,	/*   STC DC,C		; to dch */
-	0106601+CHANGE_DEV,	/*   OTB CC		; seek cmd */
-	0103701+CHANGE_DEV,	/*   STC CC,C		; to cch */
-	0102300+CHANGE_DEV,	/*   SFS DC		; addr wd ok? */
+	0106610,		/*   OTB DC		; cyl # */
+	0103710,		/*   STC DC,C		; to dch */
+	0106611,		/*   OTB CC		; seek cmd */
+	0103711,		/*   STC CC,C		; to cch */
+	0102310,		/*   SFS DC		; addr wd ok? */
 	0027710,		/*   JMP *-1		; no, wait */
 	0006400,		/*   CLB */
 	0102501,		/*   LIA 1		; read switches */
 	0002011,		/*   SLA,RSS		; <0> set? */
-	0047747,		/*   ADB BIT9		; head 2 = fixed */
-	0106600+CHANGE_DEV,	/*   OTB DC		; head/sector */
-	0103700+CHANGE_DEV,	/*   STC DC,C		; to dch */
-	0102301+CHANGE_DEV,	/*   SFS CC		; seek done? */
+	0047747,		/*   ADB BIT9		; head 2 = removable */
+	0106610,		/*   OTB DC		; head/sector */
+	0103710,		/*   STC DC,C		; to dch */
+	0102311,		/*   SFS CC		; seek done? */
 	0027720,		/*   JMP *-1		; no, wait */
 	0017757,		/*   JSB STAT		; get status */
 	0067776,		/*   LDB DMACW		; DMA control */
@@ -742,11 +829,11 @@ static const int32 dboot[IBL_LNT] = {
 	0067752,		/*   LDB CNT		; word count */
 	0106602,		/*   OTB 2 */
 	0063745,		/*   LDB RDCMD		; read cmd */
-	0102601+CHANGE_DEV,	/*   OTA CC		; to cch */
-	0103700+CHANGE_DEV,	/*   STC DC,C		; start dch */
-	0103606,		/*   STC 6,C		; start DMA */
-	0103701+CHANGE_DEV,	/*   STC CC,C		; start cch */
-	0102301+CHANGE_DEV,	/*   SFS CC		; done? */
+	0102611,		/*   OTA CC		; to cch */
+	0103710,		/*   STC DC,C		; start dch */
+	0103706,		/*   STC 6,C		; start DMA */
+	0103711,		/*   STC CC,C		; start cch */
+	0102311,		/*   SFS CC		; done? */
 	0027737,		/*   JMP *-1		; no, wait */
 	0017757,		/*   JSB STAT		; get status */
 	0027775,		/*   JMP XT		; done */
@@ -761,11 +848,11 @@ static const int32 dboot[IBL_LNT] = {
 	0, 0, 0, 0,		/* unused */
 	0000000,		/*STAT 0 */
 	0002400,		/*   CLA		; status request */
-	0102601+CHANGE_DEV,	/*   OTC CC		; to cch */
-	0103701+CHANGE_DEV,	/*   STC CC,C		; start cch */
-	0102300+CHANGE_DEV,	/*   SFS DC		; done? */
+	0102611,		/*   OTC CC		; to cch */
+	0103711,		/*   STC CC,C		; start cch */
+	0102310,		/*   SFS DC		; done? */
 	0027763,		/*   JMP *-1 */
-	0102500+CHANGE_DEV,	/*   LIA DC		; get status */
+	0102510,		/*   LIA DC		; get status */
 	0013743,		/*   AND FSMSK		; mask 15,14,3,0 */
 	0002003,		/*   SZA,RSS		; drive ready? */
 	0127757,		/*   JMP STAT,I		; yes */
@@ -774,23 +861,17 @@ static const int32 dboot[IBL_LNT] = {
 	0102030,		/*   HLT 30		; yes */
 	0027700,		/*   JMP ST		; no, retry */
 	0117751,		/*XT JSB ADDR2,I	; start program */
-	0120000+CHANGE_DEV,	/*DMACW 120000+DC */
+	0120010,		/*DMACW 120000+DC */
 	0000000 };		/*   -ST */
 
 t_stat dpc_boot (int32 unitno, DEVICE *dptr)
 {
-int32 i, dev;
+int32 dev;
 
 if (unitno != 0) return SCPE_NOFNC;			/* only unit 0 */
 dev = dpd_dib.devno;					/* get data chan dev */
-PC = ((MEMSIZE - 1) & ~IBL_MASK) & VAMASK;		/* start at mem top */
-SR = IBL_DP + (dev << IBL_V_DEV);			/* set SR */
-if (sim_switches & SWMASK ('F')) SR = SR | IBL_FIX;	/* boot from fixed? */
-for (i = 0; i < IBL_LNT; i++) {				/* copy bootstrap */
-	if (dboot[i] & CHANGE_DEV)			/* IO instr? */
-	    M[PC + i] = (dboot[i] + dev) & DMASK;
-	else M[PC + i] = dboot[i];  }
-M[PC + LDR_BASE] = (~PC + 1) & DMASK;	
+if (ibl_copy (dp_rom, dev)) return SCPE_IERR;		/* copy boot to memory */
+SR = (SR & IBL_OPT) | IBL_DP | (dev << IBL_V_DEV);	/* set SR */
+if (sim_switches & SWMASK ('R')) SR = SR | IBL_DP_REM;	/* boot from removable? */
 return SCPE_OK;
 }
-
