@@ -25,6 +25,15 @@
 
    dq		12565A 2883 disk system
 
+   07-Oct-04	JDB	Fixed enable/disable from either device
+			Shortened xtime from 5 to 3 (drive avg 156KW/second)
+			Fixed not ready/any error status
+			Fixed RAR model
+   21-Apr-04	RMS	Fixed typo in boot loader (found by Dave Bryan)
+   26-Apr-04	RMS	Fixed SFS x,C and SFC x,C
+			Fixed SR setting in IBL
+			Revised IBL loader
+			Implemented DMA SRQ (follows FLG)
    25-Apr-03	RMS	Fixed bug in status check
    10-Nov-02	RMS	Added boot command, rebuilt like 12559/13210
    09-Jan-02	WOM	Copied dp driver and mods for 2883
@@ -34,6 +43,29 @@
    - 12565 does not set error on positioner busy
    - 12565 does not set positioner busy if already on cylinder
    - 12565 does not need eoc logic, it will hit an invalid head number
+
+   The controller's "Record Address Register" (RAR) contains the CHS address of
+   the last Position or Load Address command executed.  The RAR is shared among
+   all drives on the controller.  In addition, each drive has an internal
+   position register that contains the last cylinder and head position
+   transferred to the drive during Position command execution (sector operations
+   always start with the RAR sector position).
+
+   In a real drive, the address field of the sector under the head is read and
+   compared to the RAR.  When they match, the target sector is under the head
+   and is ready for reading or writing.  If a match doesn't occur, an Address
+   Error is indicated.  In the simulator, the address field is obtained from the
+   drive's current position register during a read, i.e., the "on-disc" address
+   field is assumed to match the current position.
+
+   Reference:
+   - 12565A Disc Interface Kit Operating and Service Manual (12565-90003, Aug-1973)
+
+   The following implemented behaviors have been inferred from secondary sources
+   (diagnostics, operating system drivers, etc.), due to absent or contradictory
+   authoritative information; future correction may be needed:
+
+     1. Read Address command starts at the sector number in the RAR.
 */
 
 #include "hp2100_defs.h"
@@ -41,7 +73,7 @@
 #define UNIT_V_WLK	(UNIT_V_UF + 0)			/* write locked */
 #define UNIT_WLK	(1 << UNIT_V_WLK)
 #define FNC		u3				/* saved function */
-#define CYL		u4				/* cylinder */
+#define DRV		u4				/* drive number (DC) */
 #define UNIT_WPRT	(UNIT_WLK | UNIT_RO)		/* write prot */
 
 #define DQ_N_NUMWD	7
@@ -102,11 +134,11 @@
 #define STA_BSY		0000004				/* seeking */
 #define STA_DTE		0000002				/* data error */
 #define STA_ERR		0000001				/* any error */
-#define STA_ALLERR	(STA_NRDY + STA_EOC + STA_AER + STA_FLG + STA_DTE)
+#define STA_ANYERR	(STA_NRDY | STA_EOC | STA_AER | STA_FLG | STA_DTE)
 
 extern uint16 *M;
 extern uint32 PC, SR;
-extern uint32 dev_cmd[2], dev_ctl[2], dev_flg[2], dev_fbf[2];
+extern uint32 dev_cmd[2], dev_ctl[2], dev_flg[2], dev_fbf[2], dev_srq[2];
 extern int32 sim_switches;
 extern UNIT cpu_unit;
 
@@ -114,17 +146,19 @@ int32 dqc_busy = 0;					/* cch xfer */
 int32 dqc_cnt = 0;					/* check count */
 int32 dqc_stime = 100;					/* seek time */
 int32 dqc_ctime = 100;					/* command time */
-int32 dqc_xtime = 5;					/* xfer time */
+int32 dqc_xtime = 3;					/* xfer time */
 int32 dqc_dtime = 2;					/* dch time */
 int32 dqd_obuf = 0, dqd_ibuf = 0;			/* dch buffers */
 int32 dqc_obuf = 0;					/* cch buffers */
 int32 dqd_xfer = 0;					/* xfer in prog */
 int32 dqd_wval = 0;					/* write data valid */
 int32 dq_ptr = 0;					/* buffer ptr */
-uint8 dqc_rarc[DQ_NUMDRV] = { 0 };			/* cylinder */
-uint8 dqc_rarh[DQ_NUMDRV] = { 0 };			/* head */
-uint8 dqc_rars[DQ_NUMDRV] = { 0 };			/* sector */
-uint16 dqc_sta[DQ_NUMDRV] = { 0 };			/* status regs */
+uint8 dqc_rarc = 0;					/* RAR cylinder */
+uint8 dqc_rarh = 0;					/* RAR head */
+uint8 dqc_rars = 0;					/* RAR sector */
+uint8 dqc_ucyl[DQ_NUMDRV] = { 0 };			/* unit cylinder */
+uint8 dqc_uhed[DQ_NUMDRV] = { 0 };			/* unit head */
+uint16 dqc_sta[DQ_NUMDRV] = { 0 };			/* unit status */
 uint16 dqxb[DQ_NUMWD];					/* sector buffer */
 
 DEVICE dqd_dev, dqc_dev;
@@ -145,8 +179,8 @@ void dq_goc (int32 fnc, int32 drv, int32 time);
 */
 
 DIB dq_dib[] = {
-	{ DQD, 0, 0, 0, 0, &dqdio },
-	{ DQC, 0, 0, 0, 0, &dqcio }  };
+	{ DQD, 0, 0, 0, 0, 0, &dqdio },
+	{ DQC, 0, 0, 0, 0, 0, &dqcio }  };
 
 #define dqd_dib dq_dib[0]
 #define dqc_dib dq_dib[1]
@@ -156,14 +190,15 @@ UNIT dqd_unit = { UDATA (&dqd_svc, 0, 0) };
 REG dqd_reg[] = {
 	{ ORDATA (IBUF, dqd_ibuf, 16) },
 	{ ORDATA (OBUF, dqd_obuf, 16) },
+	{ BRDATA (DBUF, dqxb, 8, 16, DQ_NUMWD) },
+	{ DRDATA (BPTR, dq_ptr, DQ_N_NUMWD) },
 	{ FLDATA (CMD, dqd_dib.cmd, 0) },
 	{ FLDATA (CTL, dqd_dib.ctl, 0) },
 	{ FLDATA (FLG, dqd_dib.flg, 0) },
 	{ FLDATA (FBF, dqd_dib.fbf, 0) },
+	{ FLDATA (SRQ, dqd_dib.srq, 0) },
 	{ FLDATA (XFER, dqd_xfer, 0) },
 	{ FLDATA (WVAL, dqd_wval, 0) },
-	{ BRDATA (DBUF, dqxb, 8, 16, DQ_NUMWD) },
-	{ DRDATA (BPTR, dq_ptr, DQ_N_NUMWD) },
 	{ ORDATA (DEVNO, dqd_dib.devno, 6), REG_HRO },
 	{ NULL }  };
 
@@ -177,7 +212,7 @@ DEVICE dqd_dev = {
 	1, 10, DQ_N_NUMWD, 1, 8, 16,
 	NULL, NULL, &dqc_reset,
 	NULL, NULL, NULL,
-	&dqd_dib, 0 };
+	&dqd_dib, DEV_DISABLE };
 
 /* DQC data structures
 
@@ -201,16 +236,17 @@ REG dqc_reg[] = {
 	{ FLDATA (CTL, dqc_dib.ctl, 0) },
 	{ FLDATA (FLG, dqc_dib.flg, 0) },
 	{ FLDATA (FBF, dqc_dib.fbf, 0) },
-	{ BRDATA (RARC, dqc_rarc, 8, 8, DQ_NUMDRV) },
-	{ BRDATA (RARH, dqc_rarh, 8, 5, DQ_NUMDRV) },
-	{ BRDATA (RARS, dqc_rars, 8, 5, DQ_NUMDRV) },
+	{ FLDATA (SRQ, dqc_dib.srq, 0) },
+	{ DRDATA (RARC, dqc_rarc, 8), PV_RZRO },
+	{ DRDATA (RARH, dqc_rarh, 5), PV_RZRO },
+	{ DRDATA (RARS, dqc_rars, 5), PV_RZRO },
+	{ BRDATA (CYL, dqc_ucyl, 10, 8, DQ_NUMDRV), PV_RZRO },
+	{ BRDATA (HED, dqc_uhed, 10, 5, DQ_NUMDRV), PV_RZRO },
 	{ BRDATA (STA, dqc_sta, 8, 16, DQ_NUMDRV) },
 	{ DRDATA (CTIME, dqc_ctime, 24), PV_LEFT },
 	{ DRDATA (DTIME, dqc_dtime, 24), PV_LEFT },
 	{ DRDATA (STIME, dqc_stime, 24), PV_LEFT },
 	{ DRDATA (XTIME, dqc_xtime, 24), REG_NZ + PV_LEFT },
-	{ URDATA (UCYL, dqc_unit[0].CYL, 10, 8, 0,
-		  DQ_NUMDRV, PV_LEFT | REG_HRO) },
 	{ URDATA (UFNC, dqc_unit[0].FNC, 8, 8, 0,
 		  DQ_NUMDRV, REG_HRO) },
 	{ ORDATA (DEVNO, dqc_dib.devno, 6), REG_HRO },
@@ -239,14 +275,14 @@ int32 devd;
 devd = IR & I_DEVMASK;					/* get device no */
 switch (inst) {						/* case on opcode */
 case ioFLG:						/* flag clear/set */
-	if ((IR & I_HC) == 0) { setFLG (devd); }	/* STF */
+	if ((IR & I_HC) == 0) { setFSR (devd); }	/* STF */
 	break;
 case ioSFC:						/* skip flag clear */
 	if (FLG (devd) == 0) PC = (PC + 1) & VAMASK;
-	return dat;
+	break;
 case ioSFS:						/* skip flag set */
 	if (FLG (devd) != 0) PC = (PC + 1) & VAMASK;
-	return dat;
+	break;
 case ioOTX:						/* output */
 	dqd_obuf = dat;
 	if (!dqc_busy || dqd_xfer) dqd_wval = 1;	/* if !overrun, valid */
@@ -266,11 +302,11 @@ case ioCTL:						/* control clear/set */
 	    setCTL (devd);				/* set ctl, cmd */
 	    setCMD (devd);
 	    if (dqc_busy && !dqd_xfer)			/* overrun? */
-		dqc_sta[dqc_busy - 1] |= STA_DTE | STA_ERR;  }
+		dqc_sta[dqc_busy - 1] |= STA_DTE;  }
 	break;
 default:
 	break;  }
-if (IR & I_HC) { clrFLG (devd); }			/* H/C option */
+if (IR & I_HC) { clrFSR (devd); }			/* H/C option */
 return dat;
 }
 
@@ -281,14 +317,14 @@ int32 devc, fnc, drv;
 devc = IR & I_DEVMASK;					/* get device no */
 switch (inst) {						/* case on opcode */
 case ioFLG:						/* flag clear/set */
-	if ((IR & I_HC) == 0) { setFLG (devc); }	/* STF */
+	if ((IR & I_HC) == 0) { setFSR (devc); }	/* STF */
 	break;
 case ioSFC:						/* skip flag clear */
 	if (FLG (devc) == 0) PC = (PC + 1) & VAMASK;
-	return dat;
+	break;
 case ioSFS:						/* skip flag set */
 	if (FLG (devc) != 0) PC = (PC + 1) & VAMASK;
-	return dat;
+	break;
 case ioOTX:						/* output */
 	dqc_obuf = dat;
 	break;
@@ -308,8 +344,8 @@ case ioCTL:						/* control clear/set */
 	    setCTL (devc);				/* set ctl */
 	    if (!CMD (devc)) {				/* cmd clr? */
 		setCMD (devc);				/* set cmd, ctl */
-		drv = CW_GETDRV (dqc_obuf);			/* get fnc, drv */
-		fnc = CW_GETFNC (dqc_obuf);			/* from cmd word */
+		drv = CW_GETDRV (dqc_obuf);		/* get fnc, drv */
+		fnc = CW_GETFNC (dqc_obuf);		/* from cmd word */
 		switch (fnc) {				/* case on fnc */
 		case FNC_SEEK: case FNC_RCL:		/* seek, recal */
 		case FNC_CHK:				/* check */
@@ -327,7 +363,7 @@ case ioCTL:						/* control clear/set */
 	break;
 default:
 	break;  }
-if (IR & I_HC) { clrFLG (devc); }			/* H/C option */
+if (IR & I_HC) { clrFSR (devc); }			/* H/C option */
 return dat;
 }
 
@@ -335,7 +371,7 @@ return dat;
 
 void dq_god (int32 fnc, int32 drv, int32 time)
 {
-dqd_unit.CYL = drv;					/* save unit */
+dqd_unit.DRV = drv;					/* save unit */
 dqd_unit.FNC = fnc;					/* save function */
 sim_activate (&dqd_unit, time);
 return;
@@ -345,9 +381,11 @@ return;
 
 void dq_goc (int32 fnc, int32 drv, int32 time)
 {
-if (sim_is_active (&dqc_unit[drv])) {			/* still seeking? */
+int32 t;
+
+if (t = sim_is_active (&dqc_unit[drv])) {		/* still seeking? */
 	sim_cancel (&dqc_unit[drv]);			/* cancel */
-	time = time + dqc_stime;  }			/* take longer */
+	time = time + t;  }				/* include seek time */
 dqc_sta[drv] = 0;					/* clear status */
 dq_ptr = 0;						/* init buf ptr */
 dqc_busy = drv + 1;					/* set busy */
@@ -362,7 +400,7 @@ return;
    This routine handles the data channel transfers.  It also handles
    data transfers that are blocked by seek in progress.
 
-   uptr->CYL	=	target drive
+   uptr->DRV	=	target drive
    uptr->FNC	=	target function
 
    Seek substates
@@ -382,80 +420,69 @@ t_stat dqd_svc (UNIT *uptr)
 {
 int32 drv, devc, devd, st;
 
-drv = uptr->CYL;					/* get drive no */
+drv = uptr->DRV;					/* get drive no */
 devc = dqc_dib.devno;					/* get cch devno */
 devd = dqd_dib.devno;					/* get dch devno */
 switch (uptr->FNC) {					/* case function */
 
+case FNC_LA:						/* arec, need cyl */
 case FNC_SEEK:						/* seek, need cyl */
 	if (CMD (devd)) {				/* dch active? */
-	    dqc_rarc[drv] = DA_GETCYL (dqd_obuf);	/* take cyl word */
+	    dqc_rarc = DA_GETCYL (dqd_obuf);		/* set RAR from cyl word */
 	    dqd_wval = 0;				/* clr data valid */
-	    setFLG (devd);				/* set dch flg */
+	    setFSR (devd);				/* set dch flg */
 	    clrCMD (devd);				/* clr dch cmd */
-	    uptr->FNC = FNC_SEEK1;  }			/* advance state */
+	    if (uptr->FNC == FNC_LA) uptr->FNC = FNC_LA1;
+	    else uptr->FNC = FNC_SEEK1;  }		/* advance state */
 	sim_activate (uptr, dqc_xtime);			/* no, wait more */
 	break;
+
+case FNC_LA1:						/* arec, need hd/sec */
 case FNC_SEEK1:						/* seek, need hd/sec */
 	if (CMD (devd)) {				/* dch active? */
-	    dqc_rarh[drv] = DA_GETHD (dqd_obuf);	/* get head */
-	    dqc_rars[drv] = DA_GETSC (dqd_obuf);	/* get sector */
+	    dqc_rarh = DA_GETHD (dqd_obuf);		/* set RAR from head */
+	    dqc_rars = DA_GETSC (dqd_obuf);		/* set RAR from sector */
 	    dqd_wval = 0;				/* clr data valid */
-	    setFLG (devd);				/* set dch flg */
+	    setFSR (devd);				/* set dch flg */
 	    clrCMD (devd);				/* clr dch cmd */
-	    if (sim_is_active (&dqc_unit[drv])) break;	/* if busy */
-	    st = abs (dqc_rarc[drv] - dqc_unit[drv].CYL) * dqc_stime;
+	    if (uptr->FNC == FNC_LA1) {
+		setFSR (devc);				/* set cch flg */
+		clrCMD (devc);				/* clr cch cmd */
+		break;  }				/* done if Load Address */
+	    if (sim_is_active (&dqc_unit[drv])) break;	/* if busy, seek check */
+	    st = abs (dqc_rarc - dqc_ucyl[drv]) * dqc_stime;
 	    if (st == 0) st = dqc_xtime;		/* if on cyl, min time */
 	    else dqc_sta[drv] = dqc_sta[drv] | STA_BSY;	/* set busy */
+	    dqc_ucyl[drv] = dqc_rarc;			/* transfer RAR */
+	    dqc_uhed[drv] = dqc_rarh;
 	    sim_activate (&dqc_unit[drv], st);		/* schedule op */
-	    dqc_unit[drv].CYL = dqc_rarc[drv];		/* on cylinder */
 	    dqc_unit[drv].FNC = FNC_SEEK2;  }		/* advance state */
 	else sim_activate (uptr, dqc_xtime);		/* no, wait more */
 	break;
 
 case FNC_RCL:						/* recalibrate */
-	dqc_rarc[drv] = dqc_rarh[drv] = dqc_rars[drv] = 0;	/* clear RAR */
+	dqc_rarc = dqc_rarh = dqc_rars = 0;		/* clear RAR */
 	if (sim_is_active (&dqc_unit[drv])) break;	/* ignore if busy */
-	st = dqc_unit[drv].CYL * dqc_stime;		/* calc diff */
+	st = dqc_ucyl[drv] * dqc_stime;			/* calc diff */
 	if (st == 0) st = dqc_xtime;			/* if on cyl, min time */
 	else dqc_sta[drv] = dqc_sta[drv] | STA_BSY;	/* set busy */
 	sim_activate (&dqc_unit[drv], st);		/* schedule drive */
-	dqc_unit[drv].CYL = 0;				/* on cylinder */
+	dqc_ucyl[drv] = dqc_uhed[drv] = 0;		/* clear drive pos */
 	dqc_unit[drv].FNC = FNC_SEEK2;			/* advance state */
 	break;
 
-case FNC_LA:						/* arec, need cyl */
-	if (CMD (devd)) {				/* dch active? */
-	    dqc_rarc[drv] = DA_GETCYL (dqd_obuf);	/* take cyl word */
-	    dqd_wval = 0;				/* clr data valid */
-	    setFLG (devd);				/* set dch flg */
-	    clrCMD (devd);				/* clr dch cmd */
-	    uptr->FNC = FNC_LA1;  }			/* advance state */
-	sim_activate (uptr, dqc_xtime);			/* no, wait more */
-	break;
-case FNC_LA1:						/* arec, need hd/sec */
-	if (CMD (devd)) {				/* dch active? */
-	    dqc_rarh[drv] = DA_GETHD (dqd_obuf);	/* get head */
-	    dqc_rars[drv] = DA_GETSC (dqd_obuf);	/* get sector */
-	    dqd_wval = 0;				/* clr data valid */
-	    setFLG (devc);				/* set cch flg */
-	    clrCMD (devc);				/* clr cch cmd */
-	    setFLG (devd);				/* set dch flg */
-	    clrCMD (devd);  }				/* clr dch cmd */
-	else sim_activate (uptr, dqc_xtime);		/* no, wait more */
-	break;
-
 case FNC_STA:						/* read status */
 	if (CMD (devd)) {				/* dch active? */
 	    if (dqc_unit[drv].flags & UNIT_ATT)		/* attached? */
 		dqd_ibuf = dqc_sta[drv] & ~STA_DID;
 	    else dqd_ibuf = STA_NRDY;
+	    if (dqd_ibuf & STA_ANYERR)			/* errors? set flg */
+		dqd_ibuf = dqd_ibuf | STA_ERR;
 	    if (drv) dqd_ibuf = dqd_ibuf | STA_DID;
-	    setFLG (devd);				/* set dch flg */
+	    setFSR (devd);				/* set dch flg */
 	    clrCMD (devd);				/* clr dch cmd */
 	    clrCMD (devc);				/* clr cch cmd */
-	    dqc_sta[drv] = dqc_sta[drv] & 		/* clr sta flags */
-		~(STA_DTE | STA_FLG | STA_AER | STA_EOC | STA_ERR);
+	    dqc_sta[drv] = dqc_sta[drv] & ~STA_ANYERR;	/* clr sta flags */
 	    }
 	else sim_activate (uptr, dqc_xtime);		/* wait more */
 	break;
@@ -464,8 +491,6 @@ case FNC_CHK:						/* check, need cnt */
 	if (CMD (devd)) {				/* dch active? */
 	    dqc_cnt = dqd_obuf & DA_CKMASK;		/* get count */
 	    dqd_wval = 0;				/* clr data valid */
-/*	    setFLG (devd);				/* set dch flg */
-/*	    clrCMD (devd);				/* clr dch cmd */
 	    dq_goc (FNC_CHK1, drv, dqc_ctime);  }	/* sched drv */
 	else sim_activate (uptr, dqc_xtime);		/* wait more */
 	break;
@@ -504,7 +529,7 @@ drv = uptr - dqc_dev.units;				/* get drive no */
 devc = dqc_dib.devno;					/* get cch devno */
 devd = dqd_dib.devno;					/* get dch devno */
 if ((uptr->flags & UNIT_ATT) == 0) {			/* not attached? */
-	setFLG (devc);					/* set cch flg */
+	setFSR (devc);					/* set cch flg */
 	clrCMD (devc);					/* clr cch cmd */
 	dqc_sta[drv] = 0;				/* clr status */
 	dqc_busy = 0;					/* ctlr is free */
@@ -513,31 +538,29 @@ if ((uptr->flags & UNIT_ATT) == 0) {			/* not attached? */
 switch (uptr->FNC) {					/* case function */
 
 case FNC_SEEK2:						/* seek done */
-	if (uptr->CYL >= DQ_NUMCY) {			/* out of range? */
-	    dqc_sta[drv] = dqc_sta[drv] | STA_BSY | STA_ERR;
-	    dqc_unit[drv].CYL = 0;  }
+	if (dqc_ucyl[drv] >= DQ_NUMCY) {		/* out of range? */
+	    dqc_sta[drv] = dqc_sta[drv] | STA_BSY | STA_ERR;  /* seek check */
+	    dqc_ucyl[drv] = 0;  }			/* seek to cyl 0 */
 	else dqc_sta[drv] = dqc_sta[drv] & ~STA_BSY;	/* drive not busy */
 case FNC_SEEK3:
 	if (dqc_busy || FLG (devc)) {			/* ctrl busy? */
 	    uptr->FNC = FNC_SEEK3;			/* next state */
 	    sim_activate (uptr, dqc_xtime);  }		/* ctrl busy? wait */
 	else {
-	    setFLG (devc);				/* set cch flg */
+	    setFSR (devc);				/* set cch flg */
 	    clrCMD (devc);  }				/* clr cch cmd */
 	return SCPE_OK;
 
 case FNC_RA:						/* read addr */
 	if (!CMD (devd)) break;				/* dch clr? done */
-	if (dq_ptr == 0) dqd_ibuf = uptr->CYL;		/* 1st word? */
+	if (dq_ptr == 0) dqd_ibuf = dqc_ucyl[drv];	/* 1st word? */
 	else if (dq_ptr == 1) { 			/* second word? */
-	    dqd_ibuf = (dqc_rarh[drv] << DA_V_HD) |
-		(dqc_rars[drv] << DA_V_SC);
-	    dqc_rars[drv] = dqc_rars[drv] + 1;		/* incr address */
-	    if (dqc_rars[drv] >= DQ_NUMSC)		/* end of surf? */
-		dqc_rars[drv] = 0;  }
+	    dqd_ibuf = (dqc_uhed[drv] << DA_V_HD) |	/* use drive head */
+		(dqc_rars << DA_V_SC);			/* and RAR sector */
+	    dqc_rars = (dqc_rars + 1) % DQ_NUMSC;  }	/* incr sector */
 	else break;
 	dq_ptr = dq_ptr + 1;
-	setFLG (devd);					/* set dch flg */
+	setFSR (devd);					/* set dch flg */
 	clrCMD (devd);					/* clr dch cmd */
 	sim_activate (uptr, dqc_xtime);			/* sched next word */
 	return SCPE_OK;
@@ -547,18 +570,18 @@ case FNC_RD:						/* read */
 case FNC_CHK1:						/* check */
 	if (dq_ptr == 0) {				/* new sector? */
 	    if (!CMD (devd) && (uptr->FNC != FNC_CHK1)) break;
-	    if ((uptr->CYL != dqc_rarc[drv]) ||		/* wrong cyl or */
-		(dqc_rars[drv] >= DQ_NUMSC)) {		/* bad sector? */
-		dqc_sta[drv] = dqc_sta[drv] | STA_AER | STA_ERR;
+	    if ((dqc_rarc != dqc_ucyl[drv]) ||		/* RAR cyl miscompare? */
+		(dqc_rarh != dqc_uhed[drv]) ||		/* RAR head miscompare? */
+		(dqc_rars >= DQ_NUMSC)) {		/* bad sector? */
+		dqc_sta[drv] = dqc_sta[drv] | STA_AER;	/* no record found err */
 		break;  }
-	    if (dqc_rarh[drv] >= DQ_NUMSF) {		/* bad head? */
-		dqc_sta[drv] = dqc_sta[drv] | STA_EOC | STA_ERR;
+	    if (dqc_rarh >= DQ_NUMSF) {			/* bad head? */
+		dqc_sta[drv] = dqc_sta[drv] | STA_EOC;  /* end of cyl err */
 		break;  }
-	    da = GETDA (dqc_rarc[drv], dqc_rarh[drv], dqc_rars[drv]);
-	    dqc_rars[drv] = dqc_rars[drv] + 1;		/* incr address */
-	    if (dqc_rars[drv] >= DQ_NUMSC) {		/* end of surf? */
-		dqc_rars[drv] = 0;			/* wrap to */
-		dqc_rarh[drv] = dqc_rarh[drv] + 1;  }	/* next head */
+	    da = GETDA (dqc_rarc, dqc_rarh, dqc_rars);	/* calc disk addr */
+	    dqc_rars = (dqc_rars + 1) % DQ_NUMSC;	/* incr sector */
+	    if (dqc_rars == 0)				/* wrap? incr head */
+		dqc_uhed[drv] = dqc_rarh = dqc_rarh + 1;
 	    if (err = fseek (uptr->fileref, da * sizeof (int16),
 		SEEK_SET)) break;
 	    fxread (dqxb, sizeof (int16), DQ_NUMWD, uptr->fileref);
@@ -570,7 +593,7 @@ case FNC_CHK1:						/* check */
 		if (dqc_cnt == 0) break;  }		/* if zero, done */
 	    dq_ptr = 0;  }				/* wrap buf ptr */
 	if (CMD (devd) && dqd_xfer) {			/* dch on, xfer? */
-	    setFLG (devd); }				/* set flag */
+	    setFSR (devd); }				/* set flag */
 	clrCMD (devd);					/* clr dch cmd */
 	sim_activate (uptr, dqc_xtime);			/* sched next word */
 	return SCPE_OK;
@@ -579,31 +602,31 @@ case FNC_WA:						/* write address */
 case FNC_WD:						/* write */
 	if (dq_ptr == 0) {				/* sector start? */
 	    if (!CMD (devd) && !dqd_wval) break;	/* xfer done? */
-	    if(uptr->flags & UNIT_WPRT) {		/* write protect? */
-		dqc_sta[drv] = dqc_sta[drv] | STA_FLG | STA_ERR;
+	    if (uptr->flags & UNIT_WPRT) {		/* write protect? */
+		dqc_sta[drv] = dqc_sta[drv] | STA_FLG;
 		break;  }				/* done */
-	    if ((uptr->CYL != dqc_rarc[drv]) ||		/* wrong cyl or */
-		(dqc_rars[drv] >= DQ_NUMSC)) {		/* bad sector? */
-		dqc_sta[drv] = dqc_sta[drv] | STA_AER | STA_ERR;
+	    if ((dqc_rarc != dqc_ucyl[drv]) ||		/* RAR cyl miscompare? */
+		(dqc_rarh != dqc_uhed[drv]) ||		/* RAR head miscompare? */
+		(dqc_rars >= DQ_NUMSC)) {		/* bad sector? */
+		dqc_sta[drv] = dqc_sta[drv] | STA_AER;	/* no record found err */
 		break;  }
-	    if (dqc_rarh[drv] >= DQ_NUMSF) {			/* bad head? */
-		dqc_sta[drv] = dqc_sta[drv] | STA_EOC | STA_ERR;
-		break;  }  }				/* done */
+	    if (dqc_rarh >= DQ_NUMSF) {			/* bad head? */
+		dqc_sta[drv] = dqc_sta[drv] | STA_EOC;  /* end of cyl err */
+		break;  } }
 	dqxb[dq_ptr++] = dqd_wval? dqd_obuf: 0;		/* store word/fill */
 	dqd_wval = 0;					/* clr data valid */
 	if (dq_ptr >= DQ_NUMWD) {			/* buffer full? */
-	    da = GETDA (dqc_rarc[drv], dqc_rarh[drv], dqc_rars[drv]);
-	    dqc_rars[drv] = dqc_rars[drv] + 1;		/* incr address */
-	    if (dqc_rars[drv] >= DQ_NUMSC) {		/* end of surf? */
-		dqc_rars[drv] = 0;			/* wrap to */
-		dqc_rarh[drv] = dqc_rarh[drv] + 1;  }	/* next head */
+	    da = GETDA (dqc_rarc, dqc_rarh, dqc_rars);	/* calc disk addr */
+	    dqc_rars = (dqc_rars + 1) % DQ_NUMSC;	/* incr sector */
+	    if (dqc_rars == 0)				/* wrap? incr head */
+		dqc_uhed[drv] = dqc_rarh = dqc_rarh + 1;
 	    if (err = fseek (uptr->fileref, da * sizeof (int16),
 		SEEK_SET)) return TRUE;
 	    fxwrite (dqxb, sizeof (int16), DQ_NUMWD, uptr->fileref);
 	    if (err = ferror (uptr->fileref)) break;
 	    dq_ptr = 0;  }
 	if (CMD (devd) && dqd_xfer) {			/* dch on, xfer? */
-	    setFLG (devd); }				/* set flag */
+	    setFSR (devd); }				/* set flag */
 	clrCMD (devd);					/* clr dch cmd */
 	sim_activate (uptr, dqc_xtime);			/* sched next word */
 	return SCPE_OK;
@@ -611,7 +634,7 @@ case FNC_WD:						/* write */
 default:
 	return SCPE_IERR;  }				/* end case fnc */
 
-setFLG (devc);						/* set cch flg */
+setFSR (devc);						/* set cch flg */
 clrCMD (devc);						/* clr cch cmd */
 dqc_busy = 0;						/* ctlr is free */
 dqd_xfer = dqd_wval = 0;
@@ -626,24 +649,26 @@ return SCPE_OK;
 
 t_stat dqc_reset (DEVICE *dptr)
 {
-int32 i;
+int32 drv;
 
-hp_enbdis_pair (&dqc_dev, &dqd_dev);			/* make pair cons */
+hp_enbdis_pair (dptr,					/* make pair cons */
+	(dptr == &dqd_dev)? &dqc_dev: &dqd_dev);
 dqd_ibuf = dqd_obuf = 0;				/* clear buffers */
 dqc_busy = dqc_obuf = 0;
 dqd_xfer = dqd_wval = 0;
 dq_ptr = 0;
+dqc_rarc = dqc_rarh = dqc_rars = 0;			/* clear RAR */
 dqc_dib.cmd = dqd_dib.cmd = 0;				/* clear cmd */
 dqc_dib.ctl = dqd_dib.ctl = 0;				/* clear ctl */
 dqc_dib.fbf = dqd_dib.fbf = 1;				/* set fbf */
 dqc_dib.flg = dqd_dib.flg = 1;				/* set flg */
+dqc_dib.srq = dqd_dib.srq = 1;				/* srq follows flg */
 sim_cancel (&dqd_unit);					/* cancel dch */
-for (i = 0; i < DQ_NUMDRV; i++) {			/* loop thru drives */
-	sim_cancel (&dqc_unit[i]);			/* cancel activity */
-	dqc_unit[i].FNC = 0;				/* clear function */
-	dqc_unit[i].CYL = 0;
-	dqc_rarc[i] = dqc_rarh[i] = dqc_rars[i] = 0;	/* clear rar */
-	dqc_sta[i] = 0;  }
+for (drv = 0; drv < DQ_NUMDRV; drv++) {			/* loop thru drives */
+	sim_cancel (&dqc_unit[drv]);			/* cancel activity */
+	dqc_unit[drv].FNC = 0;				/* clear function */
+	dqc_ucyl[drv] = dqc_uhed[drv] = 0;		/* clear drive pos */
+	dqc_sta[drv] = 0;  }				/* clear status */
 return SCPE_OK;
 }
 
@@ -655,66 +680,81 @@ if (uptr->flags & UNIT_ATT) return SCPE_ARG;
 return SCPE_OK;
 }
 
-/* 2883/2884 bootstrap routine (subset HP 12992A ROM) */
+/* 7900/7901/2883/2884 bootstrap routine (HP 12992A ROM) */
 
-#define LDR_BASE	077
-#define CHANGE_DEV	(1 << 24)
-
-static const int32 dboot[IBL_LNT] = {
-	0106700+CHANGE_DEV,	/*ST CLC DC		; clr dch */
-	0106701+CHANGE_DEV,	/*   CLC CC		; clr cch */
-	0067771,		/*   LDA SKCMD		; seek cmd */
-	0106600+CHANGE_DEV,	/*   OTB DC		; cyl # */
-	0103700+CHANGE_DEV,	/*   STC DC,C		; to dch */
-	0106601+CHANGE_DEV,	/*   OTB CC		; seek cmd */
-	0103701+CHANGE_DEV,	/*   STC CC,C		; to cch */
-	0102300+CHANGE_DEV,	/*   SFS DC		; addr wd ok? */
-	0027707,		/*   JMP *-1		; no, wait */
+const uint16 dq_rom[IBL_LNT] = {
+	0102501,		/*ST LIA 1		; get switches */
+	0106501,		/*   LIB 1 */
+	0013765,		/*   AND D7		; isolate hd */
+	0005750,		/*   BLF,CLE,SLB */
+	0027741,		/*   JMP RD */
+	0005335,		/*   RBR,SLB,ERB	; <13>->E, set = 2883 */
+	0027717,		/*   JMP IS */
+	0102611,		/*LP OTA CC		; do 7900 status to */
+	0103711,		/*   STC CC,C		; clear first seek */
+	0102310,		/*   SFS DC */
+	0027711,		/*   JMP *-1 */
+	0002004,		/*   INA		; get next drive */
+	0053765,		/*   CPA D7		; all cleared? */
+	0002001,		/*   RSS */
+	0027707,		/*   JMP LP */
+	0067761,		/*IS LDB SEEKC		; get seek comnd */
+	0106610,		/*   OTB DC		; issue cyl addr (0) */
+	0103710,		/*   STC DC,C		; to dch */
+	0106611,		/*   OTB CC		; seek cmd */
+	0103711,		/*   STC CC,C		; to cch */
+	0102310,		/*   SFS DC		; addr wd ok? */
+	0027724,		/*   JMP *-1		; no, wait */
 	0006400,		/*   CLB */
-	0106600+CHANGE_DEV,	/*   OTB DC		; head/sector */
-	0103700+CHANGE_DEV,	/*   STC DC,C		; to dch */
-	0102301+CHANGE_DEV,	/*   SFS CC		; seek done? */
-	0027714,		/*   JMP *-1		; no, wait */
-	0063770,		/*   LDA RDCMD		; get read read */
-	0067776,		/*   LDB DMACW		; DMA control */
+	0102501,		/*   LIA 1		; get switches */
+	0002051,		/*   SEZ,SLA,RSS	; subchan = 1 or ISS */
+	0047770,		/*   ADB BIT9		; head 2 */
+	0106610,		/*   OTB DC		; head/sector */
+	0103710,		/*   STC DC,C		; to dch */
+	0102311,		/*   SFS CC		; seek done? */
+	0027734,		/*   JMP *-1		; no, wait */
+	0063731,		/*   LDA ISSRD		; get read read */
+	0002341,		/*   SEZ,CCE,RSS	; iss disc? */
+	0001100,		/*   ARS		; no, make 7900 read */
+	0067776,		/*RD LDB DMACW		; DMA control */
 	0106606,		/*   OTB 6 */
-	0067772,		/*   LDB ADDR1		; memory addr */
+	0067762,		/*   LDB ADDR1		; memory addr */
+	0077741,		/*   STB RD		; make non re-executable */
 	0106602,		/*   OTB 2 */
 	0102702,		/*   STC 2		; flip DMA ctrl */
-	0067774,		/*   LDB CNT		; word count */
+	0067764,		/*   LDB COUNT		; word count */
 	0106602,		/*   OTB 2 */
-	0102601+CHANGE_DEV,	/*   OTA CC		; to cch */
-	0103700+CHANGE_DEV,	/*   STC DC,C		; start dch */
-	0103606,		/*   STC 6,C		; start DMA */
-	0103701+CHANGE_DEV,	/*   STC CC,C		; start cch */
-	0102301+CHANGE_DEV,	/*   SFS CC		; done? */
-	0027732,		/*   JMP *-1		; no, wait */
-	0027775,		/*   JMP XT		; done */
-	0, 0, 0,		/* unused */
-	0, 0, 0, 0, 0, 0, 0, 0,
-	0, 0, 0, 0, 0, 0, 0, 0,
-	0, 0, 0, 0, 0, 0, 0, 0,
-	0020000,		/*RDCMD 020000		; read cmd */
-	0030000,		/*SKCMD 030000		; seek cmd */
+	0002041,		/*   SEZ,RSS */
+	0027766,		/*   JMP NW */
+	0102611,		/*   OTA CC		; to cch */
+	0103710,		/*   STC DC,C		; start dch */
+	0103706,		/*   STC 6,C		; start DMA */
+	0103711,		/*   STC CC,C		; start cch */
+	0037773,		/*   ISZ SK */
+	0027773,		/*   JMP SK */
+	0030000,		/*SEEKC 030000 */
 	0102011,		/*ADDR1 102011 */
 	0102055,		/*ADDR2 102055 */
-	0164000,		/*CNT   -6144. */
-	0117773,		/*XT JSB ADDR2,I	; start program */
-	0120000+CHANGE_DEV,	/*DMACW 120000+DC */
+	0164000,		/*COUNT -6144. */
+	0000007,		/*D7    7 */
+	0106710,		/*NW CLC DC		; set 'next wd is cmd' flag */
+	0001720,		/*   ALF,ALF		; move to head number loc */
+	0001000,		/*BIT9 ALS */
+	0103610,		/*   OTA DC,C		; output cold load cmd */
+	0103706,		/*   STC 6,C		; start DMA */
+	0102310,		/*   SFS DC		; done? */
+	0027773,		/*   JMP *-1		; no, wait */
+	0117763,		/*XT JSB ADDR2,I	; start program */
+	0120010,		/*DMACW 120000+DC */
 	0000000 };		/*   -ST */
 
 t_stat dqc_boot (int32 unitno, DEVICE *dptr)
 {
-int32 i, dev;
+int32 dev;
 
 if (unitno != 0) return SCPE_NOFNC;			/* only unit 0 */
 dev = dqd_dib.devno;					/* get data chan dev */
-PC = ((MEMSIZE - 1) & ~IBL_MASK) & VAMASK;		/* start at mem top */
-SR = IBL_DQ + (dev << IBL_V_DEV);			/* set SR */
-for (i = 0; i < IBL_LNT; i++) {				/* copy bootstrap */
-	if (dboot[i] & CHANGE_DEV)			/* IO instr? */
-	    M[PC + i] = (dboot[i] + dev) & DMASK;
-	else M[PC + i] = dboot[i];  }	
-M[PC + LDR_BASE] = (~PC + 1) & DMASK;
+if (ibl_copy (dq_rom, dev)) return SCPE_IERR;		/* copy boot to memory */
+SR = (SR & IBL_OPT) | IBL_DQ | (dev << IBL_V_DEV);	/* set SR */
 return SCPE_OK;
 }
