@@ -23,6 +23,10 @@
    be used in advertising or otherwise to promote the sale, use or other dealings
    in this Software without prior written authorization from Robert M Supnik.
 
+   25-Sep-04	JDB	Added memory protect device
+			Fixed display of CCA/CCB/CCE instructions
+   01-Jun-04	RMS	Added latent 13037 support
+   19-Apr-04	RMS	Recognize SFS x,C and SFC x,C
    22-Mar-02	RMS	Revised for dynamically allocated memory
    14-Feb-02	RMS	Added DMS instructions
    04-Feb-02	RMS	Fixed bugs in alter/skip display and parsing
@@ -41,6 +45,7 @@
 
 extern DEVICE cpu_dev;
 extern UNIT cpu_unit;
+extern DEVICE mp_dev;
 extern DEVICE dma0_dev, dma1_dev;
 extern DEVICE ptr_dev, ptp_dev;
 extern DEVICE tty_dev, clk_dev;
@@ -50,6 +55,7 @@ extern DEVICE msd_dev, msc_dev;
 extern DEVICE dpd_dev, dpc_dev;
 extern DEVICE dqd_dev, dqc_dev;
 extern DEVICE drd_dev, drc_dev;
+extern DEVICE ds_dev;
 extern DEVICE muxl_dev, muxu_dev, muxc_dev;
 extern DEVICE ipli_dev, iplo_dev;
 extern REG cpu_reg[];
@@ -67,12 +73,15 @@ extern uint16 *M;
 
 char sim_name[] = "HP 2100";
 
+char halt_msg[] = "HALT instruction xxxxxx";
+
 REG *sim_PC = &cpu_reg[0];
 
 int32 sim_emax = 3;
 
 DEVICE *sim_devices[] = {
 	&cpu_dev,
+	&mp_dev,
 	&dma0_dev,
 	&dma1_dev,
 	&ptr_dev,
@@ -84,6 +93,7 @@ DEVICE *sim_devices[] = {
 	&dpd_dev, &dpc_dev,
 	&dqd_dev, &dqc_dev,
 	&drd_dev, &drc_dev,
+	&ds_dev,
 	&mtd_dev, &mtc_dev,
 	&msd_dev, &msc_dev,
 	&muxl_dev, &muxu_dev, &muxc_dev,
@@ -94,7 +104,7 @@ const char *sim_stop_messages[] = {
 	"Unknown error",
 	"Unimplemented instruction",
 	"Non-existent I/O device",
-	"HALT instruction",
+	halt_msg,
 	"Breakpoint",
 	"Indirect address loop",
 	"Indirect address interrupt (should not happen!)",
@@ -231,7 +241,7 @@ static const int32 opc_val[] = {
  0105100+I_NPN, 0105120+I_NPN,
  0102101+I_NPN, 0103101+I_NPN, 0102201+I_NPC, 0102301+I_NPC,
  0102000+I_IO1, 0102100+I_IO2, 0103100+I_IO2,
- 0102200+I_IO2, 0102300+I_IO2, 0102400+I_IO1, 0106400+I_IO1,
+ 0102200+I_IO1, 0102300+I_IO1, 0102400+I_IO1, 0106400+I_IO1,
  0102500+I_IO1, 0106500+I_IO1, 0102600+I_IO1, 0106600+I_IO1,
  0102700+I_IO1, 0106700+I_IO1,
  0101710+I_NPN, 0101711+I_NPN, 0101712+I_NPN, 0101713+I_NPN,
@@ -277,8 +287,8 @@ static const char *stab[] = {
 static const int32 mtab[] = {
  0007700, 0007700, 0007700, 0007700, 0007700, 0007700, 0007700, 0007700,
  0007700, 0007700, 0007700, 0007700, 0007700, 0007700, 0007700, 0007700,
- 0006400, 0007000, 0007400, 0006400, 0007000, 0007400,
- 0002040, 0002040, 0002100, 0002200, 0002300,
+ 0007400, 0007400, 0007400, 0007400, 0007400, 0007400,
+ 0002040, 0002040, 0002300, 0002300, 0002300,
  0006020, 0006020, 0004010, 0004010,
  0006027, 0006027, 0006027, 0006027, 0006027, 0006027, 0006027, 0006027,
  0006027, 0006027, 0006027, 0006027, 0006027, 0006027, 0006027, 0006027,
