@@ -1,4 +1,4 @@
-/* vax_io.c: VAX Qbus IO simulator
+/* vax_io.c: VAX 3900 Qbus IO simulator
 
    Copyright (c) 1998-2004, Robert M Supnik
 
@@ -25,6 +25,11 @@
 
    qba		Qbus adapter
 
+   30-Sep-04	RMS	Revised Qbus interface
+			Moved mem_err, crd_err interrupts here from vax_cpu.c
+   09-Sep-04	RMS	Integrated powerup into RESET (with -p)
+   05-Sep-04	RMS	Added CRD interrupt handling
+   28-May-04	RMS	Revised I/O dispatching (from John Dundas)
    21-Mar-04	RMS	Added RXV21 support
    21-Dec-03	RMS	Fixed bug in autoconfigure vector assignment; added controls
    21-Nov-03	RMS	Added check for interrupt slot conflict (found by Dave Hittner)
@@ -98,14 +103,15 @@ int32 cq_mear = 0;					/* MEAR */
 int32 cq_sear = 0;					/* SEAR */
 int32 cq_mbr = 0;					/* MBR */
 int32 cq_ipc = 0;					/* IPC */
+int32 autcon_enb = 1;					/* autoconfig enable */
 
 extern uint32 *M;
 extern UNIT cpu_unit;
-extern int32 PSL, SISR, trpirq, mem_err, hlt_pin;
+extern int32 PSL, SISR, trpirq, mem_err, crd_err, hlt_pin;
 extern int32 p1;
 extern int32 ssc_bto;
-extern int32 autcon_enb;
 extern jmp_buf save_env;
+extern int32 sim_switches;
 extern DEVICE *sim_devices[];
 
 extern int32 ReadB (uint32 pa);
@@ -122,6 +128,10 @@ int32 eval_int (void);
 void cq_merr (int32 pa);
 void cq_serr (int32 pa);
 t_stat qba_reset (DEVICE *dptr);
+t_bool map_addr (uint32 qa, uint32 *ma);
+t_stat set_autocon (UNIT *uptr, int32 val, char *cptr, void *desc);
+t_stat show_autocon (FILE *st, UNIT *uptr, int32 val, void *desc);
+t_stat show_iospace (FILE *st, UNIT *uptr, int32 val, void *desc);
 
 /* Qbus adapter data structures
 
@@ -145,18 +155,30 @@ REG qba_reg[] = {
 	{ HRDATA (IPL16, int_req[2], 32), REG_RO },
 	{ HRDATA (IPL15, int_req[1], 32), REG_RO },
 	{ HRDATA (IPL14, int_req[0], 32), REG_RO },
+	{ FLDATA (AUTOCON, autcon_enb, 0), REG_HRO },
 	{ NULL }  };
 
+MTAB qba_mod[] = {
+	{ MTAB_XTD|MTAB_VDV|MTAB_NMO, 0, "IOSPACE", NULL,
+	  NULL, &show_iospace },
+	{ MTAB_XTD|MTAB_VDV, 1, "AUTOCONFIG", "AUTOCONFIG",
+	  &set_autocon, &show_autocon },
+	{ MTAB_XTD|MTAB_VDV, 0, NULL, "NOAUTOCONFIG",
+	  &set_autocon, NULL },
+	{ 0 }  };
+
 DEVICE qba_dev = {
-	"QBA", &qba_unit, qba_reg, NULL,
+	"QBA", &qba_unit, qba_reg, qba_mod,
 	1, 0, 0, 0, 0, 0,
 	NULL, NULL, &qba_reset,
 	NULL, NULL, NULL,
 	&qba_dib, DEV_QBUS };
 
-/* IO page addresses */
+/* IO page dispatches */
 
-DIB *dib_tab[DIB_MAX];					/* DIB table */
+static t_stat (*iodispR[IOPAGESIZE >> 1])(int32 *dat, int32 ad, int32 md);
+static t_stat (*iodispW[IOPAGESIZE >> 1])(int32 dat, int32 ad, int32 md);
+static DIB *iodibp[IOPAGESIZE >> 1];
 
 /* Interrupt request to interrupt action map */
 
@@ -174,14 +196,12 @@ int32 int_vec[IPL_HLVL][32];				/* int req to vector */
 
 int32 ReadQb (uint32 pa)
 {
-int32 i, val;
-DIB *dibp;
+int32 idx, val;
 
-for (i = 0; dibp = dib_tab[i]; i++ ) {
-	if ((pa >= dibp->ba) &&
-	   (pa < (dibp->ba + dibp->lnt))) {
-	    dibp->rd (&val, pa, READ);
-	    return val;  }  }
+idx = (pa & IOPAGEMASK) >> 1;
+if (iodispR[idx]) {
+	iodispR[idx] (&val, pa, READ);
+	return val;  }
 cq_merr (pa);
 MACH_CHECK (MCHK_READ);
 return 0;
@@ -189,14 +209,12 @@ return 0;
 
 void WriteQb (uint32 pa, int32 val, int32 mode)
 {
-int32 i;
-DIB *dibp;
+int32 idx;
 
-for (i = 0; dibp = dib_tab[i]; i++ ) {
-	if ((pa >= dibp->ba) &&
-	   (pa < (dibp->ba + dibp->lnt))) {
-	    dibp->wr (val, pa, mode);
-	    return;  }  }
+idx = (pa & IOPAGEMASK) >> 1;
+if (iodispW[idx]) {
+	iodispW[idx] (val, pa, mode);
+	return;  }
 cq_merr (pa);
 mem_err = 1;
 return;
@@ -257,6 +275,7 @@ static const int32 sw_int_mask[IPL_SMAX] = {
 
 if (hlt_pin) return IPL_HLTPIN;				/* hlt pin int */
 if ((ipl < IPL_MEMERR) && mem_err) return IPL_MEMERR;	/* mem err int */
+if ((ipl < IPL_CRDERR) && crd_err) return IPL_CRDERR;	/* crd err int */
 for (i = IPL_HMAX; i >= IPL_HMIN; i--) {		/* chk hwre int */
 	if (i <= ipl) return 0;				/* at ipl? no int */
 	if (int_req[i - IPL_HMIN]) return i;  }		/* req != 0? int */
@@ -274,6 +293,14 @@ int32 get_vector (int32 lvl)
 int32 i;
 int32 l = lvl - IPL_HMIN;
 
+if (lvl == IPL_MEMERR) {				/* mem error? */
+	mem_err = 0;
+	return SCB_MEMERR;  }
+if (lvl == IPL_CRDERR) {				/* CRD error? */
+	crd_err = 0;
+	return SCB_CRDERR;  }
+if (lvl > IPL_HMAX) {					/* error req lvl? */
+	ABORT (STOP_UIPL);  }				/* unknown intr */
 for (i = 0; int_req[l] && (i < 32); i++) {
 	if ((int_req[l] >> i) & 1) {
 	    int_req[l] = int_req[l] & ~(1u << i);
@@ -487,38 +514,37 @@ reset_all (5);						/* from qba on... */
 return;
 }
 
-/* Reset CQBIC */
-
-t_stat qba_reset (DEVICE *dptr)
-{
-int32 i;
-
-cq_scr = (cq_scr & CQSCR_BHL) | CQSCR_POK;
-cq_dser = cq_mear = cq_sear = cq_ipc = 0;
-for (i = 0; i < IPL_HLVL; i++) int_req[i] = 0;
-return SCPE_OK;
-}
-
 /* Powerup CQBIC */
 
 t_stat qba_powerup (void)
 {
 cq_mbr = 0;
 cq_scr = CQSCR_POK;
-return qba_reset (&qba_dev);
+return SCPE_OK;
+}
+
+/* Reset CQBIC */
+
+t_stat qba_reset (DEVICE *dptr)
+{
+int32 i;
+
+if (sim_switches & SWMASK ('P')) qba_powerup ();
+cq_scr = (cq_scr & CQSCR_BHL) | CQSCR_POK;
+cq_dser = cq_mear = cq_sear = cq_ipc = 0;
+for (i = 0; i < IPL_HLVL; i++) int_req[i] = 0;
+return SCPE_OK;
 }
 
-/* I/O buffer routines, aligned access
+/* Qbus I/O buffer routines, aligned access
 
-   map_ReadB	-	fetch byte buffer from memory
-   map_ReadW 	-	fetch word buffer from memory
-   map_ReadL 	-	fetch longword buffer from memory
-   map_WriteB 	-	store byte buffer into memory
-   map_WriteW 	-	store word buffer into memory
-   map_WriteL 	-	store longword buffer into memory
+   Map_ReadB	-	fetch byte buffer from memory
+   Map_ReadW 	-	fetch word buffer from memory
+   Map_WriteB 	-	store byte buffer into memory
+   Map_WriteW 	-	store word buffer into memory
 */
 
-int32 map_readB (uint32 ba, int32 bc, uint8 *buf)
+int32 Map_ReadB (uint32 ba, int32 bc, uint8 *buf)
 {
 int32 i;
 uint32 ma, dat;
@@ -545,7 +571,7 @@ else {	for (i = ma = 0; i < bc; i = i + 4, buf++) {	/* by longwords */
 return 0;
 }
 
-int32 map_readW (uint32 ba, int32 bc, uint16 *buf)
+int32 Map_ReadW (uint32 ba, int32 bc, uint16 *buf)
 {
 int32 i;
 uint32 ma,dat;
@@ -572,23 +598,7 @@ else {	for (i = ma = 0; i < bc; i = i + 4, buf++) {	/* by longwords */
 return 0;
 }
 
-int32 map_readL (uint32 ba, int32 bc, uint32 *buf)
-{
-int32 i;
-uint32 ma;
-
-ba = ba & ~03;
-bc = bc & ~03;
-for (i = ma = 0; i < bc; i = i + 4, buf++) {		/* by lw */
-	if ((ma & VA_M_OFF) == 0) {			/* need map? */
-	    if (!map_addr (ba + i, &ma) ||		/* inv or NXM? */
-		!ADDR_IS_MEM (ma)) return (bc - i);  }
-	*buf = ReadL (ma);
-	ma = ma + 4;  }
-return 0;
-}
-
-int32 map_writeB (uint32 ba, int32 bc, uint8 *buf)
+int32 Map_WriteB (uint32 ba, int32 bc, uint8 *buf)
 {
 int32 i;
 uint32 ma, dat;
@@ -615,7 +625,7 @@ else {	for (i = ma = 0; i < bc; i = i + 4, buf++) {	/* by longwords */
 return 0;
 }
 
-int32 map_writeW (uint32 ba, int32 bc, uint16 *buf)
+int32 Map_WriteW (uint32 ba, int32 bc, uint16 *buf)
 {
 int32 i;
 uint32 ma, dat;
@@ -641,22 +651,6 @@ else {	for (i = ma = 0; i < bc; i = i + 4, buf++) {	/* by longwords */
 	}
 return 0;
 }
-
-int32 map_writeL (uint32 ba, int32 bc, uint32 *buf)
-{
-int32 i;
-uint32 ma;
-
-ba = ba & ~03;
-bc = bc & ~03;
-for (i = ma = 0; i < bc; i = i + 4, buf++) {		/* by lw */
-	if ((ma & VA_M_OFF) == 0) {			/* need map? */
-	    if (!map_addr (ba + i, &ma) ||		/* inv or NXM? */
-		!ADDR_IS_MEM (ma)) return (bc - i);  }
-	WriteL (ma, *buf);
-	ma = ma + 4;  }
-return 0;
-}
 
 /* Enable/disable autoconfiguration */
 
@@ -781,114 +775,117 @@ else {	fprintf (st, "vector=%X", vec);
 return SCPE_OK;
 }
 
-/* Test for conflict in device addresses */
+/* Build dispatch tables */
 
-t_bool dev_conflict (DIB *curr)
+t_stat build_dsp_tab (DEVICE *dptr, DIB *dibp)
 {
-uint32 i, end;
-DEVICE *dptr;
-DIB *dibp;
+uint32 i, idx;
 
-end = curr->ba + curr->lnt - 1;				/* get end */
-for (i = 0; (dptr = sim_devices[i]) != NULL; i++) {	/* loop thru dev */
-	dibp = (DIB *) dptr->ctxt;			/* get DIB */
-	if ((dibp == NULL) || (dibp == curr) ||
-	    (dptr->flags & DEV_DIS)) continue;
-	if (((curr->ba >= dibp->ba) &&			/* overlap start? */
-	    (curr->ba < (dibp->ba + dibp->lnt))) ||
-	    ((end >= dibp->ba) &&			/* overlap end? */
-	    (end < (dibp->ba + dibp->lnt)))) {
-		printf ("Device %s address conflict at %08X\n",
-		    sim_dname (dptr), dibp->ba);
-		if (sim_log) fprintf (sim_log,
-		    "Device %s address conflict at %08X\n",
-		    sim_dname (dptr), dibp->ba);
-		return TRUE;  }  }
-return FALSE;
+if ((dptr == NULL) || (dibp == NULL)) return SCPE_IERR;	/* validate args */
+for (i = 0; i < dibp->lnt; i = i + 2) {			/* create entries */
+	idx = ((dibp->ba + i) & IOPAGEMASK) >> 1;	/* index into disp */
+	if ((iodispR[idx] && dibp->rd &&		/* conflict? */
+	    (iodispR[idx] != dibp->rd)) ||
+	    (iodispW[idx] && dibp->wr &&
+	    (iodispW[idx] != dibp->wr))) {
+	    printf ("Device %s address conflict at %08o\n",
+		sim_dname (dptr), dibp->ba);
+	    if (sim_log) fprintf (sim_log,
+		"Device %s address conflict at %08o\n",
+		sim_dname (dptr), dibp->ba);
+	    return SCPE_STOP;
+	    }
+	if (dibp->rd) iodispR[idx] = dibp->rd;		/* set rd dispatch */
+	if (dibp->wr) iodispW[idx] = dibp->wr;		/* set wr dispatch */
+	iodibp[idx] = dibp;				/* remember DIB */
+	}
+return SCPE_OK;
 }
 
+
 /* Build interrupt tables */
 
-t_bool build_int_vec (int32 vloc, int32 ivec, int32 (*iack)(void) )
+t_stat build_int_vec (DEVICE *dptr, DIB *dibp)
 {
-int32 ilvl = vloc / 32;
-int32 ibit = vloc % 32;
+int32 i, idx, vec, ilvl, ibit;
 
-if (iack != NULL) {
-	if (int_ack[ilvl][ibit] &&
-	   (int_ack[ilvl][ibit] != iack)) return TRUE;
-	 int_ack[ilvl][ibit] = iack;  }
-else if (ivec != 0) {
-	if (int_vec[ilvl][ibit] &&
-	    (int_vec[ilvl][ibit] != ivec)) return TRUE;
-	int_vec[ilvl][ibit] = ivec;  }
-return FALSE;
+if ((dptr == NULL) || (dibp == NULL)) return SCPE_IERR;	/* validate args */
+if (dibp->vnum > VEC_DEVMAX) return SCPE_IERR;
+for (i = 0; i < dibp->vnum; i++) {			/* loop thru vec */
+	idx = dibp->vloc + i;				/* vector index */
+	vec = dibp->vec? (dibp->vec + (i * 4)): 0;	/* vector addr */
+	ilvl = idx / 32;
+	ibit = idx % 32;
+	if ((int_ack[ilvl][ibit] && dibp->ack[i] &&	/* conflict? */
+	    (int_ack[ilvl][ibit] != dibp->ack[i])) ||
+	    (int_vec[ilvl][ibit] && vec &&
+	    (int_vec[ilvl][ibit] != vec))) {
+	    printf ("Device %s interrupt slot conflict at %d\n",
+		sim_dname (dptr), idx);
+	    if (sim_log) fprintf (sim_log,
+	 	"Device %s interrupt slot conflict at %d\n",
+		sim_dname (dptr), idx);
+	    return SCPE_STOP;
+	    }
+	if (dibp->ack[i]) int_ack[ilvl][ibit] = dibp->ack[i];
+	else if (vec) int_vec[ilvl][ibit] = vec;
+	}
+return SCPE_OK;
 }
 
 /* Build dib_tab from device list */
 
 t_stat build_dib_tab (void)
 {
-int32 i, j, k;
+int32 i, j;
 DEVICE *dptr;
 DIB *dibp;
+t_stat r;
 
 for (i = 0; i < IPL_HLVL; i++) {			/* clear int tables */
 	for (j = 0; j < 32; j++) {
 	    int_vec[i][j] = 0;
 	    int_ack[i][j] = NULL;  }  }
-for (i = j = 0; (dptr = sim_devices[i]) != NULL; i++) {	/* loop thru dev */
+for (i = 0; i < (IOPAGESIZE >> 1); i++) {		/* clear dispatch tab */
+	iodispR[i] = NULL;
+	iodispW[i] = NULL;
+	iodibp[i] = NULL;  }
+for (i = 0; (dptr = sim_devices[i]) != NULL; i++) {	/* loop thru dev */
 	dibp = (DIB *) dptr->ctxt;			/* get DIB */
 	if (dibp && !(dptr->flags & DEV_DIS)) {		/* defined, enabled? */
-	    if (dibp->vnum > VEC_DEVMAX) return SCPE_IERR;
-	    for (k = 0; k < dibp->vnum; k++) {		/* loop thru vec */
-	        if (build_int_vec (dibp->vloc + k,	/* add vector */
-		    dibp->vec + (k * 4), dibp->ack[k])) {
-		    printf ("Device %s interrupt slot conflict at %d\n",
-			sim_dname (dptr), dibp->vloc + k);
-		    if (sim_log) fprintf (sim_log,
-			"Device %s interrupt slot conflict at %d\n",
-			sim_dname (dptr), dibp->vloc + k);
-		    return SCPE_IERR;  }  }
-	    if (dibp->lnt != 0) {			/* I/O addresses? */
-		dib_tab[j++] = dibp;			/* add DIB to dib_tab */
-		if (j >= DIB_MAX) return SCPE_IERR;  }	/* too many? */	
+	    if (r = build_int_vec (dptr, dibp))		/* add to intr tab */
+		return r;
+	    if (r = build_dsp_tab (dptr, dibp))		/* add to dispatch tab */
+		return r;
 	    }						/* end if enabled */
 	}						/* end for */
-dib_tab[j] = NULL;					/* end with NULL */
-for (i = 0; (dibp = dib_tab[i]) != NULL; i++) {		/* test built dib_tab */
-	if (dev_conflict (dibp)) return SCPE_STOP;  }	/* for conflicts */
-return FALSE;
+return SCPE_OK;
 }
 
-/* Show dib_tab */
+/* Show IO space */
 
 t_stat show_iospace (FILE *st, UNIT *uptr, int32 val, void *desc)
 {
-int32 i, j, done = 0;
+uint32 i, j;
 DEVICE *dptr;
-DIB *dibt;
+DIB *dibp;
 
-build_dib_tab ();					/* build table */
-while (done == 0) {					/* sort ascending */
-	done = 1;					/* assume done */
-	for (i = 0; dib_tab[i + 1] != NULL; i++) {	/* check table */
-	    if (dib_tab[i]->ba > dib_tab[i + 1]->ba) {	/* out of order? */
-		dibt = dib_tab[i];			/* interchange */
-		dib_tab[i] = dib_tab[i + 1];
-		dib_tab[i + 1] = dibt;
-		done = 0;  }  }				/* not done */
-	}						/* end while */
-for (i = 0; dib_tab[i] != NULL; i++) {			/* print table */
-	for (j = 0, dptr = NULL; sim_devices[j] != NULL; j++) {
-	    if (((DIB*) sim_devices[j]->ctxt) == dib_tab[i]) {
-		dptr = sim_devices[j];
-		break;  }  }
-	fprintf (st, "%08X - %08X%c\t%s\n", dib_tab[i]->ba,
-		dib_tab[i]->ba + dib_tab[i]->lnt - 1,
+if (build_dib_tab ()) return SCPE_OK;			/* build IO page */
+for (i = 0, dibp = NULL; i < (IOPAGESIZE >> 1); i++) {	/* loop thru entries */
+	if (iodibp[i] && (iodibp[i] != dibp)) {		/* new block? */
+	    dibp = iodibp[i];				/* DIB for block */
+	    for (j = 0, dptr = NULL; sim_devices[j] != NULL; j++) {
+		if (((DIB*) sim_devices[j]->ctxt) == dibp) {
+		    dptr = sim_devices[j];		/* locate device */
+		    break;
+		    }					/* end if */
+		}					/* end for j */
+	fprintf (st, "%08X - %08X%c\t%s\n", dibp->ba,
+		dibp->ba + dibp->lnt - 1,
 		(dptr && (dptr->flags & DEV_FLTA))? '*': ' ',
 		dptr? sim_dname (dptr): "CPU");
-	}
+	    }						/* end if */
+	}						/* end for i */
 return SCPE_OK;
 }
 
@@ -943,7 +940,7 @@ struct auto_con auto_tab[AUTO_LNT + 1] = {
 	{ 0xf, 0x3 },					/* VS100 */
 	{ 0x3, 0x3, AUTO_DYN|AUTO_VEC, 0, IOBA_TQ, { "TQ", "TQB" } },
 	{ 0xf, 0x7 },					/* KMV11 */
-	{ 0xf, 0x7 },					/* DHU11/DHQ11 */
+	{ 0x1f, 0x7, AUTO_VEC, VH_MUXES, 0, { "VH" } },	/* DHU11/DHQ11 */
 
 	{ 0x1f, 0x7 },					/* DMZ32 */
 	{ 0x1f, 0x7 },					/* CP132 */
